import logging
import time

from dranspose.event import ResultData

logger = logging.getLogger(__name__)


class SlowReducer:
    def __init__(self, **kwargs):
        self.publish = {"map": {}}

    def process_result(self, result: ResultData, parameters=None):
        print(result)
        logger.info("processing_result %s", result)
        # if result.payload:
        time.sleep(0.5)
        logger.info("slept for 500ms")
        self.publish["map"][result.event_number] = 1
        logger.info("updated publish to %s", self.publish)

    def finish(self, parameters=None):
        print("finished dummy reducer work")



import asyncio
import logging
import pickle
from typing import Callable, Optional, Awaitable, Any, Coroutine

import aiohttp
import pytest
import uvicorn
import zmq
from pydantic_core import Url

from dranspose.controller import app
from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.parameters import ParameterList
from dranspose.protocol import (
    WorkerName,
    StreamName,
    VirtualWorker,
    VirtualConstraint,
)
from dranspose.worker import Worker, WorkerSettings
from tests.utils import wait_for_controller


@pytest.mark.asyncio
async def test_empty_params(
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[Worker], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_eiger: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
) -> None:
    config = uvicorn.Config(app, port=5000, log_level="debug")
    server = uvicorn.Server(config)
    server_task = asyncio.create_task(server.serve())
    while server.started is False:
        await asyncio.sleep(0.1)

    logging.info("started controller")
    await asyncio.sleep(1)

    await reducer("examples.params.reducer:ParamReducer")
    await create_worker(
        Worker(
            settings=WorkerSettings(
                worker_name=WorkerName("w1"),
                worker_class="examples.params.worker:ParamWorker",
            ),
        )
    )

    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )

    await asyncio.sleep(1)

    state = await wait_for_controller(streams={StreamName("eiger")})
    async with aiohttp.ClientSession() as session:
        logging.warning("ensemble state uuids are %s", state.parameters_version)

        assert state.parameters_hash is not None
        assert state.reducer is not None
        assert state.reducer.parameters_hash is not None
        assert state.parameters_hash == state.reducer.parameters_hash
        for ing in state.ingesters:
            assert state.parameters_hash == ing.parameters_hash
        for wo in state.workers:
            assert state.parameters_hash == wo.parameters_hash

        par = await session.get("http://localhost:5000/api/v1/parameters")
        assert par.status == 200
        params = ParameterList.validate_python(await par.json())

        logging.warning("params %s", params)

        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "eiger": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200

        context = zmq.asyncio.Context()

        asyncio.create_task(stream_eiger(context, 9999, ntrig - 1))

        async with aiohttp.ClientSession() as session:
            st = await session.get("http://localhost:5000/api/v1/progress")
            content = await st.json()
            while not content["finished"]:
                await asyncio.sleep(0.3)
                st = await session.get("http://localhost:5000/api/v1/progress")
                content = await st.json()

            st = await session.get("http://localhost:5001/api/v1/result/pickle")
            content = await st.content.read()
            result = pickle.loads(content)
            logging.warning("content in reducer is %s", result["params"])
            logging.warning("content in worker is %s", result["worker_params"])
            resparams = result["params"]
            for t, v in [
                (str, ""),
                (bytes, b""),
                (int, 0),
                (float, 0.0),
                (bool, False),
            ]:
                logging.info("param %s", resparams[f"{t.__name__}_param"])
                assert resparams[f"{t.__name__}_param"].value == v
            worparams = result["worker_params"]
            for t, v in [
                (str, ""),
                (bytes, b""),
                (int, 0),
                (float, 0.0),
                (bool, False),
            ]:
                logging.info("work param %s", worparams[f"{t.__name__}_param"])
                assert worparams[f"{t.__name__}_param"].value == v

        context.destroy()

    server.should_exit = True
    await server_task
    await asyncio.sleep(0.1)



import asyncio
import os
from pathlib import PosixPath
from typing import Awaitable, Callable, Coroutine, Optional, Any

import aiohttp
import zmq.asyncio

import pytest
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_eiger_legacy import (
    ZmqPullEigerLegacyIngester,
    ZmqPullEigerLegacySettings,
)
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)

from dranspose.worker import Worker, WorkerSettings
from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.asyncio
async def test_eiger_legacy(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[Worker], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_pkls: Callable[
        [zmq.Context[Any], int, os.PathLike[Any] | str, float, int],
        Coroutine[Any, Any, None],
    ],
) -> None:
    await reducer(None)
    await create_worker(
        Worker(
            settings=WorkerSettings(
                worker_name=WorkerName("w1"),
                worker_class="examples.parser.eigerlegacy:LegacyWorker",
            ),
        )
    )
    await create_ingester(
        ZmqPullEigerLegacyIngester(
            settings=ZmqPullEigerLegacySettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:22005"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("eiger")})
    async with aiohttp.ClientSession() as session:
        ntrig = 3
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "eiger": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200
        await resp.json()

    context = zmq.asyncio.Context()

    asyncio.create_task(
        stream_pkls(
            context,
            22005,
            PosixPath("tests/data/eiger-small.pkls"),
            0.001,
            zmq.PUSH,
            begin=0,  # type: ignore[call-arg]
        )
    )

    content = await wait_for_finish()

    print(content)



from typing import Optional, Literal, Any

from pydantic import BaseModel, TypeAdapter

from dranspose.protocol import ParameterName


class ParameterBase(BaseModel):
    name: ParameterName
    description: Optional[str] = None

    @staticmethod
    def to_bytes(val: Any) -> bytes:
        return str(val).encode("utf8")


class StrParameter(ParameterBase):
    dtype: Literal["str"] = "str"
    default: str = ""

    @staticmethod
    def from_bytes(by: bytes) -> str:
        try:
            return by.decode("utf8")
        except ValueError:
            return ""


class BinaryParameter(ParameterBase):
    dtype: Literal["binary"] = "binary"
    default: bytes = b""

    @staticmethod
    def from_bytes(by: bytes) -> bytes:
        return by

    @staticmethod
    def to_bytes(val: Any) -> bytes:
        return val


class IntParameter(ParameterBase):
    dtype: Literal["int"] = "int"
    default: int = 0

    @staticmethod
    def from_bytes(by: bytes) -> int:
        try:
            return int(by)
        except ValueError:
            return 0


class FloatParameter(ParameterBase):
    dtype: Literal["float"] = "float"
    default: float = 0.0

    @staticmethod
    def from_bytes(by: bytes) -> float:
        try:
            return float(by)
        except ValueError:
            return 0.0


class BoolParameter(ParameterBase):
    dtype: Literal["bool"] = "bool"
    default: bool = False

    @staticmethod
    def from_bytes(by: bytes) -> bool:
        try:
            return by == b"True"
        except ValueError:
            return False


ParameterType = (
    StrParameter | BinaryParameter | IntParameter | FloatParameter | BoolParameter
)

Parameter = TypeAdapter(ParameterType)

ParameterList = TypeAdapter(list[ParameterType])



import asyncio
import logging
from typing import Awaitable, Callable, Any, Coroutine, Optional

from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)
from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
import aiohttp

import pytest
import zmq.asyncio
import zmq
from pydantic_core import Url

from dranspose.worker import Worker, WorkerSettings
from tests.utils import wait_for_finish, wait_for_controller


async def consume_round_robin(ctx: zmq.Context[Any], ports: list[int], num: int) -> int:
    sockets = []
    for p in ports:
        s = ctx.socket(zmq.PULL)
        s.connect(f"tcp://127.0.0.1:{p}")
        sockets.append(s)

    rcvd = 0
    si = 0
    while rcvd < num:
        data = await sockets[si].recv_multipart(copy=False)
        si += 1
        si = si % len(ports)
        rcvd += 1
        logging.info("received %d th data %s", rcvd, data)

    return rcvd


@pytest.mark.asyncio
async def test_roundrobin(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName | Worker], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_eiger: Callable[
        [zmq.Context[Any], int, int, float], Coroutine[Any, Any, None]
    ],
) -> None:
    await reducer(None)
    await create_worker(
        Worker(
            settings=WorkerSettings(
                worker_name=WorkerName("w5556"),
                worker_class="examples.repub.worker:RepubWorker",
            ),
        )
    )
    await create_worker(
        Worker(
            settings=WorkerSettings(
                worker_name=WorkerName("w5557"),
                worker_class="examples.repub.worker:RepubWorker",
            ),
        )
    )
    await create_worker(
        Worker(
            settings=WorkerSettings(
                worker_name=WorkerName("w5558"),
                worker_class="examples.repub.worker:RepubWorker",
            ),
        )
    )
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("eiger")})
    async with aiohttp.ClientSession() as session:
        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "eiger": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(i % 3)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200
        await resp.json()

    context = zmq.asyncio.Context()

    collector = asyncio.create_task(
        consume_round_robin(context, [5556, 5557, 5558], ntrig - 1 + 2 * 3)
    )

    asyncio.create_task(stream_eiger(context, 9999, ntrig - 1, 0.1))

    await wait_for_finish()

    await collector

    context.destroy()



import asyncio
import logging
import pickle
from typing import Awaitable, Callable, Any, Coroutine, Optional

from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)
from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
import aiohttp

import pytest
import zmq.asyncio
import zmq
from pydantic_core import Url

from dranspose.worker import Worker, WorkerSettings
from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.asyncio
async def test_timestamps(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName | Worker], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_eiger: Callable[
        [zmq.Context[Any], int, int, float], Coroutine[Any, Any, None]
    ],
) -> None:
    await reducer("examples.timing.reducer:TimingReducer")
    await create_worker(
        Worker(
            settings=WorkerSettings(
                worker_name=WorkerName("w1"),
                worker_class="examples.timing.worker:TimingWorker",
            ),
        )
    )
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("fast")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("fast")})
    async with aiohttp.ClientSession() as session:
        st = await session.get(
            "http://localhost:5000/api/v1/load?intervals=1&intervals=10&scan=True"
        )
        load = await st.json()
        assert load == {}

        ntrig = 4
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "fast": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200
        await resp.json()

    context = zmq.asyncio.Context()

    asyncio.create_task(stream_eiger(context, 9999, ntrig - 1, 0.1))

    await wait_for_finish()

    context.destroy()

    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5001/api/v1/result/")
        content = await st.content.read()
        result = pickle.loads(content)[0]
        assert len(result["fast"]) == ntrig - 1
        for t in result["fast"]:
            assert len(t) == 2
            assert t[0] < t[1]
            if t[1] > 0.1:
                logging.warning(
                    "the pipeline end to end latency is larger then 100ms: %s", t
                )
        logging.info("content is %s", result)



import time
from typing import NewType, Any, Literal, Iterable

from pydantic import BaseModel, Field, field_validator

H5UUID = NewType("H5UUID", str)


class H5Times(BaseModel):
    created: float = Field(default_factory=time.time)
    lastModified: float = Field(default_factory=time.time)


class H5Response(H5Times):
    root: H5UUID


class H5Root(H5Response):
    owner: str = "admin"
    class_: str = Field("domain", alias="class")


class H5Group(H5Response):
    id: H5UUID
    linkCount: int = 0
    attributeCount: int = 0


class H5StrType(BaseModel):
    class_: Literal["H5T_STRING"] = Field("H5T_STRING", alias="class")
    length: Literal["H5T_VARIABLE"] = "H5T_VARIABLE"
    charSet: Literal["H5T_CSET_UTF8"] = "H5T_CSET_UTF8"
    strPad: Literal["H5T_STR_NULLTERM"] = "H5T_STR_NULLTERM"


class H5NumBase(BaseModel):
    base: str

    @field_validator("base")
    @classmethod
    def base_must_start(cls, v: str) -> str:
        if not v.startswith("H5T_"):
            raise ValueError("type base must start with H5T_")
        return v


class H5IntType(H5NumBase):
    class_: Literal["H5T_INTEGER"] = Field("H5T_INTEGER", alias="class")


class H5FloatType(H5NumBase):
    class_: Literal["H5T_FLOAT"] = Field("H5T_FLOAT", alias="class")


class H5CompType(BaseModel):
    class_: Literal["H5T_COMPOUND"] = Field("H5T_COMPOUND", alias="class")
    fields: "list[H5NamedType]"


H5Type = H5StrType | H5FloatType | H5IntType | H5CompType


class H5NamedType(BaseModel):
    name: str
    type: H5Type


class H5ScalarShape(BaseModel):
    class_: Literal["H5S_SCALAR"] = Field("H5S_SCALAR", alias="class")


class H5SimpleShape(BaseModel):
    class_: Literal["H5S_SIMPLE"] = Field("H5S_SIMPLE", alias="class")
    dims: Iterable[int]


H5Shape = H5SimpleShape | H5ScalarShape


class H5Attribute(H5Times):
    name: str
    type: H5Type
    shape: H5Shape


class H5ValuedAttribute(H5Attribute):
    value: Any


class H5Attributes(BaseModel):
    attributes: list[H5Attribute]


class H5Link(H5Times):
    class_: Literal["H5L_TYPE_HARD"] = Field("H5L_TYPE_HARD", alias="class")
    collection: Literal["datasets", "groups"]
    id: H5UUID
    title: str


class H5Dataset(H5Times):
    id: H5UUID
    attributeCount: int = 0
    shape: H5Shape
    type: H5Type
    creationProperties: dict[str, str | dict[str, str]] = {
        "allocTime": "H5D_ALLOC_TIME_LATE",
        "fillTime": "H5D_FILL_TIME_IFSET",
        "layout": {"class": "H5D_CONTIGUOUS"},
    }



import asyncio
import logging
from typing import Optional, Any

import aiohttp

from pydantic_core import Url

from dranspose.protocol import EnsembleState, StreamName, WorkerName


async def wait_for_controller(
    streams: Optional[set[StreamName]] = None,
    workers: Optional[set[WorkerName]] = None,
    controller: Url = Url("http://localhost:5000"),
) -> EnsembleState:
    async with aiohttp.ClientSession() as session:
        st = await session.get(f"{controller}api/v1/config")
        if streams is None:
            streams = set()
        if workers is None:
            workers = set()
        state = EnsembleState.model_validate(await st.json())
        timeout = 0
        while not (
            streams <= set(state.get_streams()) and workers <= set(state.get_workers())
        ):
            await asyncio.sleep(0.3)
            timeout += 1
            st = await session.get(f"{controller}api/v1/config")
            state = EnsembleState.model_validate(await st.json())
            if timeout < 20:
                logging.debug("queried for components to become available %s", state)
            elif timeout % 4 == 0:
                logging.warning("still waiting for conponents %s", state)
            elif timeout > 40:
                logging.error(
                    "streams %s and workers %s are not available", streams, workers
                )
                raise TimeoutError("failed to bring up components")

        return state


async def wait_for_finish(
    controller: Url = Url("http://localhost:5000"),
) -> dict[str, Any]:
    async with aiohttp.ClientSession() as session:
        st = await session.get(f"{controller}api/v1/progress")
        content = await st.json()
        timeout = 0
        completed_before = -1
        while not content["finished"]:
            await asyncio.sleep(0.3)
            timeout += 1
            st = await session.get(f"{controller}api/v1/progress")
            content = await st.json()
            if content["completed_events"] != completed_before:
                timeout = 0

            if timeout < 20:
                logging.debug("at progress", content)
            elif timeout % 4 == 0:
                logging.warning("stuck at progress %s", content)
            elif timeout > 40:
                logging.error("stalled at progress %s", content)
                raise TimeoutError("processing stalled")
        return content



import asyncio
import datetime
import json
import logging
import pickle
import time
from collections import defaultdict
from typing import Optional, Any

import aiohttp
import numpy as np
import pytest
import zmq
from psutil._common import snetio

from dranspose.protocol import EnsembleState, VirtualWorker, VirtualConstraint
from dranspose.workload_generator import NetworkConfig, Statistics


def get_url(path: str, svc: str = "controller") -> str:
    return f"http://dranspose-bench-{svc}.daq.maxiv.lu.se/api/v1/{path}"


def get_gen(path: str, svc: str = "large") -> str:
    return f"http://dummymax-workload-{svc}.daq.maxiv.lu.se/api/v1/{path}"


@pytest.mark.parametrize("ingester", ["gen1"])  # "fast", "large",small
@pytest.mark.parametrize("size", [32])
@pytest.mark.skipif(
    "not config.getoption('k8s')",
    reason="explicitly enable --k8s remote bench, optional --plots",
)
@pytest.mark.asyncio
async def est_single_ingester(ingester: str, plt: Any, size: int) -> None:
    async with aiohttp.ClientSession() as session:
        st = await session.get(get_url("config"))
        estate = EnsembleState.model_validate(await st.json())
        logging.info("available streams %s", estate.get_streams())

        st = await session.get(get_gen("config", svc=ingester))
        nstate = NetworkConfig.model_validate(await st.json())
        logging.info("config is %s", nstate)

        st = await session.get(get_gen("statistics", svc=ingester))
        sstate = Statistics.model_validate(await st.json())
        logging.info("stat is %s, %s", len(sstate.snapshots), sstate.snapshots[-1])

        st = await session.post(
            get_gen("open_socket", svc=ingester),
            json={"type": zmq.PUSH, "port": 9999},
        )
        state = await st.json()
        logging.info("open %s", state)

        await asyncio.sleep(1)

        ntrig = 10000
        delay = 0.000001
        resp = await session.post(
            get_url("mapping"),
            json={
                ingester: [
                    [
                        VirtualWorker(constraint=VirtualConstraint(i // 10)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200
        uuid = await resp.json()

        logging.info("mapping uuid is %s", uuid)

        await asyncio.sleep(0.4)

        st = await session.post(
            get_gen("frames", svc=ingester),
            json={"number": ntrig, "time": delay, "shape": [size, size]},
        )
        state = await st.json()
        logging.info("sending frames %s", state)

        worker_times = defaultdict(list)
        ingester_times = defaultdict(list)
        measured_times = []

        exp_start: Optional[float] = None

        st = await session.get(get_url("progress"))
        content = await st.json()
        while not content["finished"]:
            await asyncio.sleep(0.8)
            st = await session.get(get_url("progress"))
            content = await st.json()
            logging.info("progress %s", content)

            st = await session.get(get_url("config"))
            if exp_start is None:
                exp_start = time.time()
            measured_times.append(time.time() - exp_start)
            conf = EnsembleState.model_validate(await st.json())
            msg = []
            for wk in conf.workers:
                worker_times[wk.name].append(wk.event_rate)
            for ik in conf.ingesters:
                ingester_times[ik.name].append(ik.event_rate)
            for k in conf.workers + conf.ingesters + [conf.reducer]:
                if k is not None:
                    msg.append(f"{k.name}:{k.processed_events} -- {k.event_rate}")
            logging.info("state is \n%s", "\n".join(msg))

        st = await session.post(get_gen("close_socket", svc=ingester))
        state = await st.json()
        logging.info("close %s", state)

        st = await session.get(get_gen("statistics", svc=ingester))
        stat = Statistics.model_validate_json(await st.read())
        last = {}
        last_sent = 0
        for t, sent, stats in stat.snapshots:
            if sent - last_sent == 0:
                continue
            logging.info(
                "timestamp %s: sent %d, Δ%d",
                datetime.datetime.fromtimestamp(t),
                sent,
                sent - last_sent,
            )
            last_sent = sent
            for ifname, nst in stats.items():
                if ifname not in last:
                    last[ifname] = nst
                    logging.info("    %s: %s", ifname, nst)
                else:
                    delta = snetio(*map(lambda x: x[1] - x[0], zip(last[ifname], nst)))
                    logging.info("   Δ%s: %s", ifname, delta)
                    last[ifname] = nst

        if exp_start is None:
            return

        bytes_times = []
        sent_times = []
        stat_times = []
        ifname = "purple0"
        for i in range(len(stat.snapshots) - 1, 0, -1):
            t, sent, stats = stat.snapshots[i]
            tp, sentp, statsp = stat.snapshots[i - 1]
            stat_times.append(t - exp_start)
            bytes_times.append(
                snetio(
                    *map(lambda x: x[1] - x[0], zip(statsp[ifname], stats[ifname]))
                ).bytes_sent
                / (t - tp)
                / 1024
                / 1024
            )
            sent_times.append((sent - sentp) / (t - tp))
            if t < exp_start and sent - sentp == 0:
                break

        plt.figure(figsize=(10, 10))

        for w in worker_times:
            plt.plot(measured_times, worker_times[w], label=w)
        plt.plot(
            measured_times,
            list(map(sum, zip(*worker_times.values()))),
            label="worker sum",
        )
        for dname in ingester_times:
            plt.plot(measured_times, ingester_times[dname], label=dname)
        plt.plot(stat_times, sent_times, label="generated packets")
        plt.plot(stat_times, bytes_times, label="outgoing MiBi/s")
        plt.title(f"{ntrig} frames, {delay}s delay, {size * size * 2} bytes/f")
        plt.ylabel("pakets/s")
        plt.xlabel("time in s")
        plt.legend(loc="upper center", bbox_to_anchor=(0.5, -0.1))


@pytest.mark.parametrize("size", [1024])
@pytest.mark.skipif(
    "not config.getoption('k8s')",
    reason="explicitly enable --k8s remote bench, optional --plots",
)
@pytest.mark.asyncio
async def est_single_latency(plt: Any, size: int) -> None:
    async with aiohttp.ClientSession() as session:
        ingester = "fast"
        st = await session.get(get_url("config"))
        estate = EnsembleState.model_validate(await st.json())
        logging.info("available streams %s", estate.get_streams())

        st = await session.get(get_gen("config", svc=ingester))
        nstate = NetworkConfig.model_validate(await st.json())
        logging.info("config is %s", nstate)

        st = await session.get(get_gen("statistics", svc=ingester))
        sstate = Statistics.model_validate(await st.json())
        logging.info("stat is %s, %s", len(sstate.snapshots), sstate.snapshots[-1])

        st = await session.post(
            get_gen("open_socket", svc=ingester),
            json={"type": zmq.PUSH, "port": 9999},
        )
        state = await st.json()
        logging.info("open %s", state)

        await asyncio.sleep(1)

        ntrig = 300
        delay = 0.05
        resp = await session.post(
            get_url("mapping"),
            json={
                ingester: [
                    [
                        VirtualWorker(constraint=VirtualConstraint(i // 10)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200
        uuid = await resp.json()

        logging.info("mapping uuid is %s", uuid)

        await asyncio.sleep(0.4)

        st = await session.post(
            get_gen("frames", svc=ingester),
            json={"number": ntrig, "time": delay, "shape": [size, size]},
        )
        state = await st.json()
        logging.info("sending frames %s", state)

        st = await session.get(get_url("progress"))
        content = await st.json()
        while not content["finished"]:
            await asyncio.sleep(0.8)
            st = await session.get(get_url("progress"))
            content = await st.json()
            logging.info("progress %s", content)

        st = await session.post(get_gen("close_socket", svc=ingester))
        state = await st.json()
        logging.info("close %s", state)

        st = await session.get(get_url("result/$", svc="reducer"))
        content = await st.content.read()
        result = pickle.loads(content)
        logging.info("times are %s", result)


@pytest.mark.parametrize("size", [2])
@pytest.mark.skipif(
    "not config.getoption('k8s')",
    reason="explicitly enable --k8s remote bench, optional --plots",
)
@pytest.mark.asyncio
async def est_dual_ingester(plt: Any, size: int) -> None:
    async with aiohttp.ClientSession() as session:
        ingesters = ["fast", "large"]
        st = await session.get(get_url("config"))
        state = EnsembleState.model_validate(await st.json())
        logging.info("available streams %s", state.get_streams())

        for ing in ingesters:
            await session.post(
                get_gen("open_socket", svc=ing),
                json={"type": zmq.PUSH, "port": 9999},
            )

        await asyncio.sleep(1)

        ntrig = 100000
        delay = 0.00000001
        await session.post(
            get_url("mapping"),
            json={
                ing: [
                    [
                        VirtualWorker(constraint=VirtualConstraint(i // 10)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ]
                for ing in ingesters
            },
        )

        await asyncio.sleep(0.4)

        for ing in ingesters:
            await session.post(
                get_gen("frames", svc=ing),
                json={"number": ntrig, "time": delay, "shape": [size, size]},
            )
        logging.info("sending frames %s", state)

        worker_times = defaultdict(list)
        ingester_times = defaultdict(list)
        measured_times = []

        exp_start = None

        st = await session.get(get_url("progress"))
        content = await st.json()
        while not content["finished"]:
            await asyncio.sleep(0.8)
            st = await session.get(get_url("progress"))
            content = await st.json()
            logging.info("progress %s", content)

            st = await session.get(get_url("config"))
            if exp_start is None:
                exp_start = time.time()
            measured_times.append(time.time() - exp_start)
            conf = EnsembleState.model_validate(await st.json())
            msg = []
            for wk in conf.workers:
                worker_times[wk.name].append(wk.event_rate)
            for ik in conf.ingesters:
                ingester_times[ik.name].append(ik.event_rate)
            for k in conf.workers + conf.ingesters + [conf.reducer]:
                if k is not None:
                    msg.append(f"{k.name}:{k.processed_events} -- {k.event_rate}")
            logging.info("state is \n%s", "\n".join(msg))

        st = await session.post(get_gen("close_socket", svc=ingesters[0]))
        st = await session.post(get_gen("close_socket", svc=ingesters[1]))
        state = await st.json()
        logging.info("close %s", state)

        if exp_start is None:
            return

        bytes_times = defaultdict(list)
        sent_times = defaultdict(list)
        stat_times = defaultdict(list)

        for ing in ingesters:
            st = await session.get(get_gen("statistics", svc=ing))
            stat = Statistics.model_validate_json(await st.read())

            for i in range(len(stat.snapshots) - 1, 0, -1):
                t, sent, stats = stat.snapshots[i]
                tp, sentp, statsp = stat.snapshots[i - 1]
                stat_times[ing].append(t - exp_start)
                bytes_times[ing].append(
                    snetio(
                        *map(
                            lambda x: x[1] - x[0],
                            zip(statsp["purple0"], stats["purple0"]),
                        )
                    ).bytes_sent
                    / (t - tp)
                    / 1024
                    / 1024
                )
                sent_times[ing].append((sent - sentp) / (t - tp))
                if t < exp_start and sent - sentp == 0:
                    break

        plt.figure(figsize=(10, 10))

        for w in worker_times:
            plt.plot(measured_times, worker_times[w], label=w)
        plt.plot(
            measured_times,
            list(map(sum, zip(*worker_times.values()))),
            label="worker sum",
        )
        for dname in ingester_times:
            plt.plot(measured_times, ingester_times[dname], label=dname)
        for ing in ingesters:
            plt.plot(stat_times[ing], sent_times[ing], label="generated packets")
            plt.plot(stat_times[ing], bytes_times[ing], label="outgoing MiBi/s")
        plt.title(f"{ntrig} frames, {delay}s delay, {size * size * 2} bytes/f")
        plt.ylabel("pakets/s")
        plt.xlabel("time in s")
        plt.legend(loc="upper center", bbox_to_anchor=(0.5, -0.1))


@pytest.mark.parametrize("size", [256])
@pytest.mark.parametrize("num_ing", [4, 8])
@pytest.mark.skipif(
    "not config.getoption('k8s')",
    reason="explicitly enable --k8s remote bench, optional --plots",
)
@pytest.mark.asyncio
async def est_multi_ingester(plt: Any, size: int, num_ing: int) -> None:
    async with aiohttp.ClientSession() as session:
        ingesters = [f"gen{i}" for i in range(num_ing)]
        st = await session.get(get_url("config"))
        state = EnsembleState.model_validate(await st.json())
        logging.info("available streams %s", state.get_streams())

        for ing in ingesters:
            logging.info(
                "opening socket %s with %s", ing, get_gen("open_socket", svc=ing)
            )
            await session.post(
                get_gen("open_socket", svc=ing),
                json={"type": zmq.PUSH, "port": 9999},
            )

        await asyncio.sleep(1)

        ntrig = 100000
        delay = 0.00000001
        await session.post(
            get_url("mapping"),
            json={
                ing: [
                    [
                        VirtualWorker(
                            constraint=VirtualConstraint(i // 100)
                        ).model_dump(mode="json")
                    ]
                    for i in range(1, ntrig)
                ]
                for ing in ingesters
            },
        )

        await asyncio.sleep(0.4)

        for ing in ingesters:
            await session.post(
                get_gen("frames", svc=ing),
                json={"number": ntrig, "time": delay, "shape": [size, size]},
            )
        logging.info("sending frames %s", state)

        worker_times = defaultdict(list)
        ingester_times = defaultdict(list)
        measured_times = []

        exp_start = None

        st = await session.get(get_url("progress"))
        content = await st.json()
        while not content["finished"]:
            await asyncio.sleep(0.8)
            st = await session.get(get_url("progress"))
            content = await st.json()
            logging.info("progress %s", content)

            st = await session.get(get_url("config"))
            if exp_start is None:
                exp_start = time.time()
            measured_times.append(time.time() - exp_start)
            conf = EnsembleState.model_validate(await st.json())
            msg = []
            for wk in conf.workers:
                worker_times[wk.name].append(wk.event_rate)
            for ik in conf.ingesters:
                ingester_times[ik.name].append(ik.event_rate)
            for k in conf.workers + conf.ingesters + [conf.reducer]:
                if k is not None:
                    msg.append(f"{k.name}:{k.processed_events} -- {k.event_rate}")
            logging.info("state is \n%s", "\n".join(msg))

        for ing in ingesters:
            st = await session.post(get_gen("close_socket", svc=ing))
            state = await st.json()
            logging.info("close %s", state)

        if exp_start is None:
            return

        bytes_times = defaultdict(list)
        sent_times = defaultdict(list)
        stat_times = defaultdict(list)

        for ing in ingesters:
            st = await session.get(get_gen("statistics", svc=ing))
            stat = Statistics.model_validate_json(await st.read())

            for i in range(len(stat.snapshots) - 1, 0, -1):
                t, sent, stats = stat.snapshots[i]
                tp, sentp, statsp = stat.snapshots[i - 1]
                stat_times[ing].append(t - exp_start)
                bytes_times[ing].append(
                    snetio(
                        *map(
                            lambda x: x[1] - x[0],
                            zip(statsp["purple0"], stats["purple0"]),
                        )
                    ).bytes_sent
                    / (t - tp)
                    / 1024
                    / 1024
                )
                sent_times[ing].append((sent - sentp) / (t - tp))
                if t < exp_start and sent - sentp == 0:
                    break

        plt.figure(figsize=(10, 10 + num_ing))

        for w in worker_times:
            plt.plot(measured_times, worker_times[w], label=w)
        plt.plot(
            measured_times,
            list(map(sum, zip(*worker_times.values()))),
            label="worker sum",
        )
        for dname in ingester_times:
            # plt.plot(measured_times, ingester_times[dname], label=dname)
            pass
        for ing in ingesters:
            # plt.plot(stat_times[ing], sent_times[ing], label="generated packets")
            plt.plot(stat_times[ing], bytes_times[ing], label="outgoing MiBi/s")
        plt.title(f"{ntrig} frames, {delay}s delay, {size * size * 2} bytes/f")
        plt.ylabel("pakets/s")
        plt.xlabel("time in s")
        plt.legend(loc="upper center", bbox_to_anchor=(0.5, -0.1))


@pytest.mark.parametrize("size", [32])  # 1024, 512, 256, 128, 64, 32, 16, 8
@pytest.mark.skipif(
    "not config.getoption('k8s')",
    reason="explicitly enable --k8s remote bench, optional --plots",
)
@pytest.mark.asyncio
async def test_raw_zmq(plt: Any, size: int) -> None:
    async with aiohttp.ClientSession() as session:
        exp_start = time.time()

        maxfps = 20000
        if size < 256:
            nframes = maxfps * 10
        else:
            nframes = 10000

        await session.post(
            get_gen("open_socket", svc="gen0"),
            json={"type": zmq.PUSH, "port": 9999},
        )
        await asyncio.sleep(1)

        st = await session.post(
            get_gen("connect_socket", svc="gen1"),
            json={"type": zmq.PULL, "url": "tcp://172.18.11.224:9999"},
        )
        state = await st.json()
        logging.info("open %s", state)

        st = await session.post(
            get_gen("frames", svc="gen0"),
            json={"number": nframes, "time": 0.0000001, "shape": [size, size]},
        )
        state = await st.json()
        logging.info("sending frames %s", state)

        st = await session.get(get_gen("finished", svc="gen0"))
        state = await st.json()
        logging.info("is finished %s", state)
        while not state:
            await asyncio.sleep(0.5)
            st = await session.get(get_gen("finished", svc="gen0"))
            state = await st.json()

            st = await session.get(get_gen("statistics", svc="gen0"))
            stat = await st.json()
            logging.info("sender fps %s", stat["fps"])

            st = await session.get(get_gen("statistics", svc="gen1"))
            stat = await st.json()
            logging.info("receiver fps %s", stat["fps"])

        st = await session.post(get_gen("close_socket", svc="gen0"))
        state = await st.json()
        logging.info("close source %s", state)

        st = await session.post(get_gen("close_socket", svc="gen1"))
        state = await st.json()
        logging.info("close sink %s", state)

        bytes_tx_times = defaultdict(list)
        bytes_rx_times = defaultdict(list)
        sent_times = defaultdict(list)
        stat_times = defaultdict(list)

        ingesters = ["gen0", "gen1"]
        for ing in ingesters:
            st = await session.get(get_gen("statistics", svc=ing))
            stat = Statistics.model_validate_json(await st.read())

            if len(stat.deltas) > 0:
                logging.info(
                    "receiver latencies mean %f, max %f",
                    np.mean(stat.deltas),
                    max(stat.deltas),
                )

            ifname = "purple0"
            for i in range(len(stat.snapshots) - 1, 0, -1):
                t, sent, stats = stat.snapshots[i]
                tp, sentp, statsp = stat.snapshots[i - 1]
                stat_times[ing].append(t - exp_start)
                bytes_tx_times[ing].append(
                    snetio(
                        *map(lambda x: x[1] - x[0], zip(statsp[ifname], stats[ifname]))
                    ).bytes_sent
                    / (t - tp)
                    / 1024
                    / 1024
                )
                bytes_rx_times[ing].append(
                    snetio(
                        *map(lambda x: x[1] - x[0], zip(statsp[ifname], stats[ifname]))
                    ).bytes_recv
                    / (t - tp)
                    / 1024
                    / 1024
                )
                sent_times[ing].append((sent - sentp) / (t - tp))
                if t < exp_start and sent - sentp == 0:
                    break

        with open(f"bench/raw_zmq_{size * size * 2}.json", "w") as f:
            json.dump(
                {
                    "times": stat_times,
                    "packets": sent_times,
                    "tx": bytes_tx_times,
                    "rx": bytes_rx_times,
                },
                f,
            )
        for ing in ingesters:
            plt.plot(stat_times[ing], sent_times[ing], label=f"generated packets {ing}")
            plt.plot(
                stat_times[ing], bytes_tx_times[ing], label=f"outgoing MiBi/s {ing}"
            )
            plt.plot(
                stat_times[ing], bytes_rx_times[ing], label=f"incoming MiBi/s {ing}"
            )
        plt.title(f"{nframes} frames, no delay, {size * size * 2 / 1024} kbytes/f")
        plt.ylabel("pakets/s")
        plt.xlabel("time in s")
        plt.legend(loc="upper center", bbox_to_anchor=(0.5, -0.1))



import json
from types import UnionType

import numpy as np
import zmq

from dranspose.data.stream1 import Stream1Packet, Stream1Data
from dranspose.event import StreamData


def parse(data: StreamData) -> UnionType:
    """
    Parses a stream1 packet, which returns a start,image or end message

    Arguments:
        data: a frame comming from a stream1 source

    Returns:
        a Stream1Packet
    """
    assert data.typ == "STINS"
    frame = data.frames[0]
    if isinstance(frame, zmq.Frame):
        val = json.loads(frame.bytes)
    elif isinstance(frame, bytes):
        val = json.loads(frame)
    else:
        raise Exception("invalid StreamData")

    packet = Stream1Packet.validate_python(val)
    if isinstance(packet, Stream1Data):
        assert data.length == 2
        if packet.compression == "none":
            bufframe = data.frames[1]
            if isinstance(bufframe, zmq.Frame):
                bufframe = bufframe.bytes
            buf = np.frombuffer(bufframe, dtype=packet.type)
            img = buf.reshape(packet.shape)
            packet.data = img

    return packet



import asyncio
import logging
from typing import Callable, Any, Coroutine

import pytest
import zmq.asyncio

from dranspose.helpers.utils import cancel_and_wait


@pytest.mark.asyncio
async def test_zmq_cancel(
    stream_eiger: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]]
) -> None:
    async def consume() -> None:
        c = zmq.asyncio.Context()
        s = c.socket(zmq.PULL)
        s.connect("tcp://localhost:9999")
        for i in range(13):
            logging.info("recv %d", i)
            task = s.poll()
            waiting = asyncio.gather(*[task])
            if i == 6:
                await cancel_and_wait(waiting)
                continue
            done = await waiting
            logging.info("done %s", done)
            data = s.recv_multipart(copy=False)
            logging.info("i %d, data %s", i, data)
        c.destroy()

    ctask = asyncio.create_task(consume())

    context = zmq.asyncio.Context()
    await stream_eiger(context, 9999, 4)
    await asyncio.sleep(1)
    await stream_eiger(context, 9999, 4)
    await ctask

    context.destroy()



import asyncio
from asyncio import StreamWriter
from typing import AsyncGenerator, Optional

import zmq
from pydantic_core import Url

from dranspose.event import StreamData
from dranspose.ingester import Ingester, IngesterSettings
from dranspose.protocol import StreamName


class TcpPcapSettings(IngesterSettings):
    upstream_url: Url


class TcpPcapIngester(Ingester):
    """
    A simple ingester class to comsume a stream from the streaming-receiver repub port
    """

    def __init__(self, settings: Optional[TcpPcapSettings] = None) -> None:
        if settings is not None:
            self._pcap_settings = settings
        else:
            self._pcap_settings = TcpPcapSettings()

        super().__init__(settings=self._pcap_settings)
        self.writer: Optional[StreamWriter] = None

    async def run_source(self, stream: StreamName) -> AsyncGenerator[StreamData, None]:
        reader, self.writer = await asyncio.open_connection(
            self._pcap_settings.upstream_url.host, self._pcap_settings.upstream_url.port
        )
        self.writer.write(b"\n")
        data = await reader.readline()
        assert data == b"OK\n"
        stop = False
        self._logger.info("connected to %s", self._pcap_settings.upstream_url)

        while True:
            while True:
                self._logger.debug("clear up insocket")
                line = await reader.readline()
                self._logger.debug("received line %s", line)
                if line == b"":
                    stop = True
                    break
                if not line.startswith(b"arm_time: "):
                    self._logger.debug("discard non-start packet")
                else:
                    self._logger.info("start of new sequence %s", line)
                    header = [line]
                    while line := await reader.readline():
                        header.append(line)
                        if line.strip() == b"":
                            break
                    self._logger.info("use header %s", header)
                    frame = zmq.Frame(b"".join(header))
                    yield StreamData(typ="PCAP_RAW", frames=[frame])
                    break
            if stop:
                break
            while True:
                line = await reader.readline()
                self._logger.debug("send frame %s", line)
                frame = zmq.Frame(line)
                yield StreamData(typ="PCAP_RAW", frames=[frame])
                if line.startswith(b"END"):
                    break

        while True:
            self._logger.debug("discarding messages until next run")
            line = await reader.readline()
            if line == b"":
                break

    async def stop_source(self, stream: StreamName) -> None:
        if self.writer:
            self.writer.close()
            await self.writer.wait_closed()
            self.writer = None



"""
This is the central service to orchestrate all distributed components

"""
import asyncio
import json
import os.path
from asyncio import Task
from collections import defaultdict
from types import UnionType
from typing import Any, AsyncGenerator, Optional, Annotated, Literal

import logging
import time

import cbor2
from pydantic import UUID4, ValidationError
from starlette.requests import Request
from starlette.responses import Response, FileResponse

from dranspose.distributed import DistributedSettings
from dranspose.helpers.utils import parameters_hash, done_callback, cancel_and_wait
from dranspose.mapping import MappingSequence, Map
import redis.asyncio as redis
import redis.exceptions as rexceptions
from importlib.metadata import version

from contextlib import asynccontextmanager
from fastapi import FastAPI, HTTPException, Query
from fastapi.middleware.cors import CORSMiddleware

from dranspose.parameters import (
    Parameter,
    ParameterType,
    StrParameter,
)
from dranspose.protocol import (
    IngesterState,
    WorkerState,
    RedisKeys,
    EnsembleState,
    WorkerUpdate,
    DistributedStateEnum,
    WorkerName,
    StreamName,
    EventNumber,
    VirtualWorker,
    WorkParameter,
    ControllerUpdate,
    ReducerState,
    WorkerTimes,
    HashDigest,
    ParameterName,
    SystemLoadType,
    IntervalLoad,
    WorkerLoad,
    DistributedUpdate,
    ReducerUpdate,
    IngesterUpdate,
    WorkAssignmentList,
    WorkAssignment,
    IngesterName,
    MappingName,
)

logger = logging.getLogger(__name__)


class ControllerSettings(DistributedSettings):
    pass


class Controller:
    def __init__(self, settings: ControllerSettings | None = None):
        self.settings = settings
        if self.settings is None:
            self.settings = ControllerSettings()

        self.redis = redis.from_url(
            f"{self.settings.redis_dsn}?decode_responses=True&protocol=3"
        )
        self.mapping = MappingSequence(parts={}, sequence=[])
        self.mapping_update_lock = asyncio.Lock()
        self.parameters: dict[ParameterName, WorkParameter] = {}
        self.parameters_hash = parameters_hash(self.parameters)
        self.completed: dict[EventNumber, list[WorkerName]] = defaultdict(list)
        self.to_reduce: set[tuple[EventNumber, WorkerName]] = set()
        self.reduced: dict[EventNumber, list[WorkerName]] = defaultdict(list)
        self.processed_event_no: int = 0
        self.completed_events: list[int] = []
        self.finished_components: list[UnionType] = []
        self.external_stop = False
        self.assign_task: Task[None]
        self.config_fetch_time: float = 0
        self.config_cache: EnsembleState | None = None
        self.default_task: Task[None]
        self.consistent_task: Task[None]
        self.worker_timing: dict[
            WorkerName, dict[EventNumber, WorkerTimes]
        ] = defaultdict(dict)
        self.start_time: float

    async def run(self) -> None:
        logger.debug("started controller run")
        dist_lock = await self.redis.set(
            RedisKeys.lock(),
            "🔒",
            ex=10,
            nx=True,
        )
        logger.debug("result of lock acquisition %s", dist_lock)
        while dist_lock is None:
            logger.warning("another controller is already running, will retry ")
            await asyncio.sleep(2)
            dist_lock = await self.redis.set(
                RedisKeys.lock(),
                "🔒",
                ex=10,
                nx=True,
            )
        logger.info("controller lock acquired")

        self.assign_task = asyncio.create_task(self.assign_work())
        self.assign_task.add_done_callback(done_callback)
        self.default_task = asyncio.create_task(self.default_parameters())
        self.default_task.add_done_callback(done_callback)
        self.consistent_task = asyncio.create_task(self.consistent_parameters())
        self.consistent_task.add_done_callback(done_callback)
        self.lock_task = asyncio.create_task(self.hold_lock())
        self.lock_task.add_done_callback(done_callback)

    async def hold_lock(self) -> None:
        while True:
            await asyncio.sleep(7)
            dist_lock = await self.redis.set(
                RedisKeys.lock(),
                "🔒",
                ex=10,
                xx=True,
            )
            if dist_lock is False:
                logger.warning("The lock was lost")

    async def get_configs(self) -> EnsembleState:
        if time.time() - self.config_fetch_time < 0.5 and self.config_cache is not None:
            return self.config_cache
        async with self.redis.pipeline() as pipe:
            await pipe.keys(RedisKeys.config("ingester"))
            await pipe.keys(RedisKeys.config("worker"))
            ingester_keys, worker_keys = await pipe.execute()
        async with self.redis.pipeline() as pipe:
            await pipe.mget(ingester_keys)
            await pipe.mget(worker_keys)
            await pipe.get(RedisKeys.config("reducer"))
            ingester_json, worker_json, reducer_json = await pipe.execute()

        ingesters = []
        for i in ingester_json:
            try:
                ingester = IngesterState.model_validate_json(i)
                ingesters.append(ingester)
            except ValidationError:
                pass
        workers = []
        for w in worker_json:
            try:
                worker = WorkerState.model_validate_json(w)
                workers.append(worker)
            except ValidationError:
                pass
        reducer = None
        if reducer_json:
            reducer = ReducerState.model_validate_json(reducer_json)
        dranspose_version = version("dranspose")

        parameters_version: dict[ParameterName, UUID4] = {
            n: p.uuid for n, p in self.parameters.items()
        }
        redis_param_keys = await self.redis.keys(RedisKeys.parameters("*", "*"))
        logger.debug("redis param keys are %s", redis_param_keys)
        self.config_cache = EnsembleState(
            ingesters=ingesters,
            workers=workers,
            reducer=reducer,
            controller_version=dranspose_version,
            parameters_version=parameters_version,
            parameters_hash=self.parameters_hash,
        )
        self.config_fetch_time = time.time()
        return self.config_cache

    async def get_load(self, intervals: list[int], scan: bool = True) -> SystemLoadType:
        ret = {}
        for wn, wt in self.worker_timing.items():
            last_event = max(wt.keys())
            itval: dict[int | Literal["scan"], IntervalLoad] = {}
            for interval in intervals:
                evs = list(
                    filter(
                        lambda x: x >= len(self.completed_events) - interval,
                        wt.keys(),
                    )
                )
                itval[interval] = IntervalLoad(
                    total=sum([wt[e].total for e in evs]),
                    active=sum([wt[e].active for e in evs]),
                    events=len(evs),
                )
            if scan:
                evs = list(wt.keys())
                itval["scan"] = IntervalLoad(
                    total=sum([wt[e].total for e in evs]),
                    active=sum([wt[e].active for e in evs]),
                    events=len(evs),
                )
            ret[wn] = WorkerLoad(last_event=last_event, intervals=itval)
        return ret

    async def dump_map_and_parameters(self) -> None:
        """Saves all current parameters to a JSON/bin file if dump_prefix is defined."""
        dump_prefix = self.parameters[ParameterName("dump_prefix")].data.decode("utf8")
        if len(dump_prefix) > 0:
            parts = {
                str(n): m.model_dump(mode="json") for n, m in self.mapping.parts.items()
            }
            mapping_json = {
                "parts": parts,
                "sequence": self.mapping.sequence,
            }
            filename = f"{dump_prefix}mapping-{self.mapping.uuid}.json"
            try:
                with open(filename, "w") as f:
                    json.dump(mapping_json, f)
                logger.info(f"Mapping saved to {filename}")
            except Exception as e:
                logger.error(f"Could not write mapping dump {e}")
            filename = f"parameters-{self.mapping.uuid}"
            bin_file = False
            pars = []
            for name, par in self.parameters.items():
                if b"\x00" in par.data:
                    bin_file = True
                    pars.append({"name": name, "data": par.data})
                else:
                    pars.append({"name": name, "data": par.data.decode("utf8")})
            try:
                if bin_file:
                    filename = f"{dump_prefix}{filename}.cbor"
                    with open(filename, "wb") as f:
                        cbor2.dump(pars, f)
                else:
                    filename = f"{dump_prefix}{filename}.json"
                    with open(filename, "w") as f:
                        json.dump(pars, f)
                logger.info(f"Parameters saved to {filename}")
            except Exception as e:
                logger.error(f"Could not write parameters dump {e}")

    async def set_mapping(self, m: MappingSequence) -> None:
        async with self.mapping_update_lock:
            logger.debug("cancelling assign task")
            await cancel_and_wait(self.assign_task)
            logger.debug(
                "deleting keys %s and %s",
                RedisKeys.ready(self.mapping.uuid),
                RedisKeys.assigned(self.mapping.uuid),
            )
            await self.redis.delete(RedisKeys.ready(self.mapping.uuid))
            await self.redis.delete(RedisKeys.assigned(self.mapping.uuid))

            # cleaned up
            self.mapping = m
            cupd = ControllerUpdate(
                mapping_uuid=self.mapping.uuid,
                parameters_version={n: p.uuid for n, p in self.parameters.items()},
                active_streams=list(m.all_streams),
            )
            logger.debug("send controller update %s", cupd)
            await self.redis.xadd(
                RedisKeys.updates(),
                {"data": cupd.model_dump_json()},
            )

            cfgs = await self.get_configs()
            timeout = 0
            while cfgs.reducer is None or set(
                [u.mapping_uuid for u in cfgs.ingesters]
                + [u.mapping_uuid for u in cfgs.workers]
                + [cfgs.reducer.mapping_uuid]
            ) != {self.mapping.uuid}:
                await asyncio.sleep(0.1)
                timeout += 1
                cfgs = await self.get_configs()
                if timeout < 50:
                    logger.debug("updated configs %s", cfgs)
                elif timeout % 10 == 0:
                    logger.warning(
                        "still waiting for configs %s to reach everyone", cfgs
                    )
                elif timeout > 100:
                    logger.error("could not distribute mapping within 10 seconds")
                    raise TimeoutError("could not distribute mapping")
            logger.info("new mapping with uuid %s distributed", self.mapping.uuid)
            if ParameterName("dump_prefix") in self.parameters:
                await self.dump_map_and_parameters()
            self.assign_task = asyncio.create_task(self.assign_work())
            self.assign_task.add_done_callback(done_callback)

    async def set_param(self, name: ParameterName, data: bytes) -> HashDigest:
        """Writes a parameter in the controller's store, updates the controller's
        hash, and updating the parameter hash. It also sends a controller update
        to the Redis stream to notify other components of the new parameter."""

        param = WorkParameter(name=name, data=data)
        logger.debug("distributing parameter %s with uuid %s", param.name, param.uuid)
        await self.redis.set(RedisKeys.parameters(name, param.uuid), param.data)
        self.parameters[name] = param
        logger.debug("stored parameter %s locally", name)
        self.parameters_hash = parameters_hash(self.parameters)
        logger.debug("parameter hash is now %s", self.parameters_hash)
        cupd = ControllerUpdate(
            mapping_uuid=self.mapping.uuid,
            parameters_version={n: p.uuid for n, p in self.parameters.items()},
        )
        logger.debug("send update %s", cupd)
        await self.redis.xadd(
            RedisKeys.updates(),
            {"data": cupd.model_dump_json()},
        )
        return self.parameters_hash

    async def describe_parameters(self) -> list[ParameterType]:
        desc_keys = await self.redis.keys(RedisKeys.parameter_description())
        param_json = await self.redis.mget(desc_keys)

        params: list[ParameterType] = []
        for i in param_json:
            val: ParameterType = Parameter.validate_json(i)
            params.append(val)

        params.append(
            StrParameter(
                name=ParameterName("dump_prefix"),
                description="Prefix to dump ingester values",
            )
        )
        return sorted(params, key=lambda x: x.name)

    async def default_parameters(self) -> None:
        """The descriptions (and default values) for the parameters are defined
        in the worker's payload and published to redis. default_parameters
        retieves the description and adds it to the list of parameters if
        they are missing."""
        while True:
            try:
                desc_keys = await self.redis.keys(RedisKeys.parameter_description())
                param_json = await self.redis.mget(desc_keys)
                for i in param_json:
                    val: ParameterType = Parameter.validate_json(i)
                    if val.name not in self.parameters:
                        logger.info(
                            "set parameter %s to default %s, (type %s)",
                            val.name,
                            val.default,
                            val.__class__,
                        )
                        await self.set_param(
                            val.name, val.__class__.to_bytes(val.default)
                        )
                await asyncio.sleep(2)
            except asyncio.exceptions.CancelledError:
                break

    async def consistent_parameters(self) -> None:
        while True:
            try:
                # make sure self.parameters is present in redis

                key_names = []
                for name, param in self.parameters.items():
                    key_names.append(RedisKeys.parameters(name, param.uuid))
                if len(key_names) > 0:
                    ex_key_no = await self.redis.exists(*key_names)
                    logger.debug(
                        "check for param values in redis, %d exist of %d: %s",
                        ex_key_no,
                        len(key_names),
                        key_names,
                    )
                    if ex_key_no < len(key_names):
                        logger.warning(
                            "the redis parameters don't match the controller parameters, rewriting"
                        )
                        async with self.redis.pipeline() as pipe:
                            for name, param in self.parameters.items():
                                await pipe.set(
                                    RedisKeys.parameters(name, param.uuid), param.data
                                )
                            await pipe.execute()

                consistent = []
                cfg = await self.get_configs()
                if cfg.reducer and cfg.reducer.parameters_hash != self.parameters_hash:
                    consistent.append(("reducer", cfg.reducer.parameters_hash))
                for wo in cfg.workers:
                    if wo.parameters_hash != self.parameters_hash:
                        consistent.append((wo.name, wo.parameters_hash))
                for ing in cfg.ingesters:
                    if ing.parameters_hash != self.parameters_hash:
                        consistent.append((ing.name, ing.parameters_hash))

                if len(consistent) > 0:
                    logger.info(
                        "inconsistent parameters %s, redistribute hash %s",
                        consistent,
                        self.parameters_hash,
                    )
                    cupd = ControllerUpdate(
                        mapping_uuid=self.mapping.uuid,
                        parameters_version={
                            n: p.uuid for n, p in self.parameters.items()
                        },
                        target_parameters_hash=self.parameters_hash,
                    )
                    logger.debug("send consistency update %s", cupd)
                    await self.redis.xadd(
                        RedisKeys.updates(),
                        {"data": cupd.model_dump_json()},
                    )
                await asyncio.sleep(2)
            except asyncio.exceptions.CancelledError:
                break

    async def _update_processing_times(self, update: WorkerUpdate) -> None:
        if update.completed is None:
            return
        if update.processing_times:
            self.worker_timing[update.worker][
                update.completed[-1]
            ] = update.processing_times
        for compev, has_result in zip(update.completed, update.has_result):
            self.completed[compev].append(update.worker)
            logger.debug("added completed to set %s", self.completed)
            wa = self.mapping.get_event_workers(compev)
            if wa.get_all_workers() == set(self.completed[compev]):
                self.completed_events.append(compev)
            if has_result:
                toadd = True
                if compev in self.reduced:
                    logger.debug(
                        "events %s already received from reducer", self.reduced
                    )
                    if update.worker in self.reduced[compev]:
                        # process worker update very late and already got reduced
                        logger.debug(
                            "event %s from worker %s was already reduced",
                            compev,
                            update.worker,
                        )
                        toadd = False
                if toadd:
                    logger.debug(
                        "waiting for reduction for event %s from worker %s",
                        compev,
                        update.worker,
                    )
                    self.to_reduce.add((compev, update.worker))
        # logger.error("time processtime %s", time.perf_counter() - start)

    async def assign_worker_in_mapping(
        self, worker: WorkerName, completed: EventNumber
    ) -> None:
        cfg = await self.get_configs()
        # logger.error("time cfg %s", time.perf_counter() - start)
        logger.debug(
            "assigning worker %s with all %s",
            worker,
            [ws.name for ws in cfg.workers],
        )
        virt = self.mapping.assign_next(
            next(w for w in cfg.workers if w.name == worker),
            cfg.workers,
            completed=completed,
            horizon=5,
        )
        # logger.error("time assign %s", time.perf_counter() - start)
        logger.debug("assigned worker %s to %s", worker, virt)
        logger.debug(
            "send out complete events in range(%d, %d)",
            self.processed_event_no,
            self.mapping.complete_events,
        )
        assignments: list[WorkAssignment] = []
        for evn in range(self.processed_event_no, self.mapping.complete_events):
            wrks = self.mapping.get_event_workers(EventNumber(evn))
            assignments.append(wrks)
            if wrks.get_all_workers() == set():
                self.completed[wrks.event_number] = []
                self.completed_events.append(wrks.event_number)
            if evn % 1000 == 0:
                logger.info(
                    "1000 events in %lf",
                    time.perf_counter() - self.start_time,
                )
                self.start_time = time.perf_counter()
        if len(assignments) > 0:
            logger.debug("send out assignment %s", assignments)
            await self.redis.xadd(
                RedisKeys.assigned(self.mapping.uuid),
                {"data": WorkAssignmentList.dump_json(assignments)},
            )

        # logger.error("time sent out %s", time.perf_counter() - start)
        self.processed_event_no = self.mapping.complete_events

    async def _process_worker_update(self, update: WorkerUpdate) -> None:
        logger.debug("got a ready worker %s", update)
        if update.state == DistributedStateEnum.READY:
            await self.assign_worker_in_mapping(update.worker, EventNumber(0))
        if update.state == DistributedStateEnum.IDLE:
            # start = time.perf_counter()
            await self.assign_worker_in_mapping(update.worker, update.completed[-1])

            asyncio.create_task(self._update_processing_times(update))

    async def completed_finish(self) -> bool:
        fin_workers: set[WorkerName] = set()
        reducer = False
        fin_ingesters: set[IngesterName] = set()
        for upd in self.finished_components:
            if isinstance(upd, WorkerUpdate):
                fin_workers.add(upd.worker)
            elif isinstance(upd, ReducerUpdate):
                reducer = True
            elif isinstance(upd, IngesterUpdate):
                fin_ingesters.add(upd.ingester)

        cfgs = await self.get_configs()

        return (
            set([w.name for w in cfgs.workers]) == fin_workers
            and set([i.name for i in cfgs.ingesters]) == fin_ingesters
            and reducer
        )

    async def assign_work(self) -> None:
        last = 0
        self.external_stop = False
        self.processed_event_no = 0
        self.completed = defaultdict(list)
        self.reduced = defaultdict(list)
        self.completed_events = []
        self.to_reduce = set()
        self.finished_components = []
        self.worker_timing = defaultdict(dict)
        notify_finish = True
        self.start_time = time.perf_counter()
        while True:
            try:
                workers = await self.redis.xread(
                    {RedisKeys.ready(self.mapping.uuid): last},
                    block=1000,
                )
                if RedisKeys.ready(self.mapping.uuid) in workers:
                    for ready in workers[RedisKeys.ready(self.mapping.uuid)][0]:
                        logger.debug("got ready raw data: %s", ready)
                        update = DistributedUpdate.validate_json(ready[1]["data"])
                        if isinstance(update, WorkerUpdate):
                            await self._process_worker_update(update)
                        elif isinstance(update, ReducerUpdate):
                            if (
                                update.completed is not None
                                and update.worker is not None
                            ):
                                compev = update.completed
                                self.reduced[compev].append(update.worker)
                                logger.debug("added reduced to set %s", self.reduced)
                                if (compev, update.worker) in self.to_reduce:
                                    self.to_reduce.remove((compev, update.worker))
                        elif isinstance(update, IngesterUpdate):
                            pass

                        if hasattr(update, "state"):
                            if update.state == DistributedStateEnum.FINISHED:
                                logger.info("distributed %s has finished", update)
                                self.finished_components.append(update)

                        last = ready[0]
                logger.debug(
                    "checking if finished, completed %d, len %d, to_reduce %s",
                    len(self.completed_events),
                    self.mapping.len(),
                    self.to_reduce,
                )
                if (
                    (
                        len(self.completed_events) > 0
                        and len(self.completed_events) == self.mapping.len()
                        and len(self.to_reduce) == 0
                    )
                    or self.external_stop
                ) and notify_finish:
                    # all events done, send close
                    cupd = ControllerUpdate(
                        mapping_uuid=self.mapping.uuid,
                        parameters_version={
                            n: p.uuid for n, p in self.parameters.items()
                        },
                        finished=True,
                    )
                    logger.debug("send finished update %s", cupd)
                    await self.redis.xadd(
                        RedisKeys.updates(),
                        {"data": cupd.model_dump_json()},
                    )
                    notify_finish = False
            except rexceptions.ConnectionError:
                break

    async def close(self) -> None:
        await cancel_and_wait(self.default_task)
        await cancel_and_wait(self.consistent_task)
        await self.redis.delete(RedisKeys.updates())
        queues = await self.redis.keys(RedisKeys.ready("*"))
        if len(queues) > 0:
            await self.redis.delete(*queues)
        assigned = await self.redis.keys(RedisKeys.assigned("*"))
        if len(assigned) > 0:
            await self.redis.delete(*assigned)
        params = await self.redis.keys(RedisKeys.parameters("*", "*"))
        if len(params) > 0:
            await self.redis.delete(*params)
        await cancel_and_wait(self.lock_task)
        await self.redis.delete(RedisKeys.lock())
        await self.redis.aclose()


ctrl: Controller


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
    # Load the ML model
    global ctrl
    ctrl = Controller()
    # run_task = asyncio.create_task(ctrl.run())
    await ctrl.run()
    # run_task.add_done_callback(done_callback)
    yield
    # await cancel_and_wait(run_task)
    await ctrl.close()
    # Clean up the ML models and release the resources


app = FastAPI(lifespan=lifespan)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
)


@app.get("/")
async def read_index() -> FileResponse:
    return FileResponse(
        os.path.join(os.path.dirname(__file__), "frontend", "index.html")
    )


@app.get("/api/v1/config")
async def get_configs() -> EnsembleState:
    global ctrl
    return await ctrl.get_configs()


@app.get("/api/v1/status")
async def get_status() -> dict[str, Any]:
    global ctrl
    return {
        "work_completed": ctrl.completed,
        "last_assigned": ctrl.mapping.complete_events,
        "assignment": [am.assignments for am in ctrl.mapping.active_maps],
        "completed_events": ctrl.completed_events,
        "finished": await ctrl.completed_finish(),
        "processing_times": ctrl.worker_timing,
    }


@app.get("/api/v1/load")
async def get_load(
    intervals: Annotated[list[int] | None, Query()] = None, scan: bool = True
) -> SystemLoadType:
    global ctrl
    if intervals is None:
        intervals = [1, 10]
    return await ctrl.get_load(intervals, scan)


@app.get("/api/v1/progress")
async def get_progress() -> dict[str, Any]:
    global ctrl
    return {
        "last_assigned": ctrl.mapping.complete_events,
        "completed_events": len(ctrl.completed_events),
        "total_events": ctrl.mapping.len(),
        "finished": await ctrl.completed_finish(),
    }


@app.post("/api/v1/mapping")
async def set_mapping(
    mapping: dict[StreamName, list[Optional[list[VirtualWorker]]]],
    all_wrap: bool = True,
) -> UUID4 | str:
    global ctrl
    config = await ctrl.get_configs()
    if set(mapping.keys()) - set(config.get_streams()) != set():
        logger.warning(
            "bad request streams %s not available",
            set(mapping.keys()) - set(config.get_streams()),
        )
        raise HTTPException(
            status_code=400,
            detail=f"streams {set(mapping.keys()) - set(config.get_streams())} not available",
        )
    m = MappingSequence(
        parts={MappingName("main"): mapping},
        sequence=[MappingName("main")],
        add_start_end=all_wrap,
    )
    if len(config.workers) < m.min_workers():
        logger.warning(
            "only %d workers available, but %d required",
            len(config.workers),
            m.min_workers(),
        )
        raise HTTPException(
            status_code=400,
            detail=f"only {len(config.workers)} workers available, but {m.min_workers()} required",
        )
    await ctrl.set_mapping(m)
    return m.uuid


@app.get("/api/v1/mapping")
async def get_mapping() -> dict[str, Any]:
    global ctrl
    return {"parts": ctrl.mapping.parts, "sequence": ctrl.mapping.sequence}


@app.post("/api/v1/sequence")
async def set_sequence(
    parts: dict[MappingName, dict[StreamName, list[Optional[list[VirtualWorker]]]]],
    sequence: list[MappingName],
    all_wrap: bool = True,
) -> UUID4 | str:
    global ctrl
    config = await ctrl.get_configs()
    m = MappingSequence(
        parts=parts,
        sequence=sequence,
        add_start_end=all_wrap,
    )
    if m.all_streams - set(config.get_streams()) != set():
        logger.warning(
            "bad request streams %s not available",
            m.all_streams - set(config.get_streams()),
        )
        raise HTTPException(
            status_code=400,
            detail=f"streams {m.all_streams - set(config.get_streams())} not available",
        )

    if len(config.workers) < m.min_workers():
        logger.warning(
            "only %d workers available, but %d required",
            len(config.workers),
            m.min_workers(),
        )
        raise HTTPException(
            status_code=400,
            detail=f"only {len(config.workers)} workers available, but {m.min_workers()} required",
        )
    await ctrl.set_mapping(m)
    return m.uuid


@app.post("/api/v1/stop")
async def stop() -> None:
    global ctrl
    logger.info("externally stopped scan")
    ctrl.external_stop = True


@app.get("/api/v1/logs")
async def get_logs(level: str = "INFO") -> Any:
    global ctrl
    data = await ctrl.redis.xrange("dranspose_logs", "-", "+")
    logs = []
    # TODO: once python 3.10 support is dropped, use
    # levels = logging.getLevelNamesMapping()
    levels = {
        level: logging.__getattribute__(level)
        for level in dir(logging)
        if level.isupper() and isinstance(logging.__getattribute__(level), int)
    }
    minlevel = levels.get(level.upper(), logging.INFO)
    for entry in data:
        msglevel = levels[entry[1].get("levelname", "DEBUG")]
        if msglevel >= minlevel:
            logs.append(entry[1])
    return logs


@app.post("/api/v1/sardana_hook")
async def set_sardana_hook(
    info: dict[Literal["streams"] | Literal["scan"], Any]
) -> UUID4 | str:
    global ctrl
    config = await ctrl.get_configs()
    print(info)
    if "scan" not in info:
        return "no scan info"
    if "nb_points" not in info["scan"]:
        return "no nb_points in scan"
    if "streams" not in info:
        return "streams required"
    for st in set(config.get_streams()).intersection(set(info["streams"])):
        print("use stream", st)
    logger.debug("create new mapping")
    m = Map.from_uniform(
        set(config.get_streams()).intersection(set(info["streams"])),
        info["scan"]["nb_points"],
    )
    s = MappingSequence(
        parts={MappingName("sardana"): m.mapping},
        sequence=[MappingName("sardana")],
        add_start_end=True,
    )
    logger.debug("set mapping")
    await ctrl.set_mapping(s)
    return s.uuid


@app.post("/api/v1/parameter/{name}")
async def post_param(request: Request, name: ParameterName) -> HashDigest:
    data = await request.body()
    logger.info("got %s: %s (len %d)", name, data[:100], len(data))
    u = await ctrl.set_param(name, data)
    return u


@app.get("/api/v1/parameter/{name}")
async def get_param(name: ParameterName) -> Response:
    if name not in ctrl.parameters:
        raise HTTPException(status_code=404, detail="Parameter not found")

    data = ctrl.parameters[name].data
    return Response(data, media_type="application/x.bytes")


@app.get("/api/v1/parameters")
async def param_descr() -> list[ParameterType]:
    global ctrl
    return await ctrl.describe_parameters()



import asyncio
import logging
from typing import Any

import aiohttp
import numpy as np
import pytest
import uvicorn
from fastapi import FastAPI
from dranspose.helpers.h5dict import router

import h5pyd


@pytest.mark.asyncio
async def test_mapping() -> None:
    app = FastAPI()

    app.include_router(router, prefix="/results")

    def get_data() -> dict[str, Any]:
        return {
            "map": {
                "x": [
                    np.float64(-14.96431272),
                    np.float64(-14.76401095),
                    np.float64(-14.56276384),
                    np.float64(-14.361908),
                    np.float64(-14.16112703),
                ],
                "y": [
                    np.float64(-10.00637736),
                    np.float64(-10.00502708),
                    np.float64(-10.00403313),
                    np.float64(-10.00349819),
                    np.float64(-10.00320074),
                ],
                "values": [
                    np.float32(0.6831444),
                    np.float32(0.0),
                    np.float32(0.039953336),
                    np.float32(0.14946304),
                    np.float32(0.0),
                ],
            },
            "control": {},
            "azint": {"data": []},
        }

    app.state.get_data = get_data

    config = uvicorn.Config(app, port=5000, log_level="debug")
    server = uvicorn.Server(config)
    server_task = asyncio.create_task(server.serve())
    while server.started is False:
        await asyncio.sleep(0.1)

    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5000/results")
        data = await st.json()
        logging.info("content %s", data)

    def work() -> None:
        f = h5pyd.File("/", "r", endpoint="http://localhost:5000/results")
        logging.info("file %s", list(f.keys()))
        logging.info("typ %s", f["map"])
        logging.info("comp %s", f["map/values"])
        logging.info("comp data %s", f["map/values"][:])
        assert f["map/values"].dtype == np.float32
        assert f["map/x"].dtype == np.float64

    loop = asyncio.get_event_loop()
    await loop.run_in_executor(None, work)

    server.should_exit = True
    await server_task

    await asyncio.sleep(0.5)


@pytest.mark.asyncio
async def test_root() -> None:
    app = FastAPI()

    app.include_router(router, prefix="/results")

    def get_data() -> dict[str, Any]:
        dt = np.dtype({"names": ["a", "b"], "formats": [float, int]})

        arr = np.array([(0.5, 1)], dtype=dt)

        return {
            "live": 34,
            "other": {"third": [1, 2, 3]},  # only _attrs allowed in root
            "other_attrs": {"NX_class": "NXother"},
            "spaced group": {"space ds": 4, "space ds_attrs": {"bla": 5}},
            "spaced group_attrs": {"spaced attr": 3},
            "image": np.ones((1000, 1000)),
            "specialtyp": np.ones((10, 10), dtype=">u8"),
            "specialtyp_attrs": {"NXdata": "NXspecial"},
            "composite": arr,
            "composite_attrs": {"axes": ["I", "q"]},
            "hello": "World",
            "_attrs": {"NX_class": "NXentry"},
        }

    app.state.get_data = get_data

    config = uvicorn.Config(app, port=5000, log_level="debug")
    server = uvicorn.Server(config)
    server_task = asyncio.create_task(server.serve())
    while server.started is False:
        await asyncio.sleep(0.1)

    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5000/results")
        data = await st.json()
        logging.info("content %s", data)

        comr_data = await session.get(
            "http://localhost:5000/results/datasets/d-h5dict-2F696D616765/value"
        )
        data_len = comr_data.content_length
        logging.info("data len %d", data_len)
        logging.info("headers are %s", comr_data.headers.items())
        assert comr_data.headers["Content-Encoding"] == "gzip"
        assert data_len == 11683

    async with aiohttp.ClientSession() as session:
        comr_data = await session.get(
            "http://localhost:5000/results/datasets/d-h5dict-2F696D616765/value",
            headers={"Accept-Encoding": "deflate"},
        )
        data_len = comr_data.content_length
        logging.info("data len %d", data_len)
        logging.info("headers are %s", comr_data.headers.items())
        assert "Content-Encoding" not in comr_data.headers
        assert data_len == 8e6

    def work() -> None:
        f = h5pyd.File("/", "r", endpoint="http://localhost:5000/results")
        logging.info("file %s", f["live"][()])
        logging.info("typ %s", f["specialtyp"])
        logging.info("comp %s", f["composite"])
        logging.info("comp data %s", f["composite"][:])
        logging.info("spaces %s", list(f["spaced group"].keys()))
        assert list(f["spaced group"].keys()) == ["space ds"]
        assert f["spaced group"].attrs["spaced attr"] == 3
        assert f["spaced group/space ds"][()] == 4
        assert f["spaced group/space ds"].attrs["bla"] == 5
        assert f["specialtyp"].dtype == ">u8"
        assert f["specialtyp"].attrs["NXdata"] == "NXspecial"
        assert f["other"].attrs["NX_class"] == "NXother"
        assert f.attrs["NX_class"] == "NXentry"
        assert list(f["composite"].attrs["axes"]) == ["I", "q"]
        assert f["composite"].dtype == np.dtype(
            {"names": ["a", "b"], "formats": [float, int]}
        )
        assert np.array_equal(f["image"], np.ones((1000, 1000)))

    loop = asyncio.get_event_loop()
    await loop.run_in_executor(None, work)

    server.should_exit = True
    await server_task

    await asyncio.sleep(0.5)


@pytest.mark.asyncio
async def test_slice() -> None:
    app = FastAPI()

    app.include_router(router, prefix="/results")

    def get_data() -> dict[str, Any]:
        return {
            "image": [[r * 100 + c for c in range(10)] for r in range(10)],
        }

    app.state.get_data = get_data

    config = uvicorn.Config(app, port=5000, log_level="debug")
    server = uvicorn.Server(config)
    server_task = asyncio.create_task(server.serve())
    while server.started is False:
        await asyncio.sleep(0.1)

    async with aiohttp.ClientSession() as session:
        comr_data = await session.get(
            "http://localhost:5000/results/datasets/d-h5dict-2F696D616765/value?select=[2:5,7:20]",
            headers={"Accept-Encoding": "deflate"},
        )
        orig = np.array([[r * 100 + c for c in range(10)] for r in range(10)])
        logging.info("original %s", orig.dtype)
        raw = await comr_data.content.read()
        arr = np.frombuffer(raw, dtype=np.int64)
        assert np.array_equal(orig[2:5, 7:20].reshape(-1), arr)
        logging.info("got data %s", arr)

    server.should_exit = True
    await server_task

    await asyncio.sleep(0.5)


def test_dtype_to_h5() -> None:
    numpytypes = {
        "u1",
        ">u2",
        ">u4",
        ">u8",
        "u1",
        "<u2",
        "<u4",
        "<u8",
        "i1",
        ">i2",
        ">i4",
        ">i8",
        "<i2",
        "<i4",
        "<i8",
        # ">f2",
        ">f4",
        ">f8",
        # ">f16",
        # "<f2",
        "<f4",
        "<f8",
        # "<f16"
    }
    for typ in numpytypes:
        dtype = np.dtype(typ)
        # {"base": "H5T_IEEE_F64LE", "class": "H5T_FLOAT"}
        # {"base": "H5T_STD_I64LE", "class": "H5T_INTEGER"}
        canonical = dtype.descr[0][1]
        if canonical.startswith(">"):
            order = "BE"
        else:
            order = "LE"
        bytelen = int(canonical[2:])
        if canonical[1] == "f":
            # floating
            htyp = {"class": "H5T_FLOAT"}
            htyp["base"] = f"H5T_IEEE_F{8 * bytelen}{order}"
        elif canonical[1] in ["u", "i"]:
            htyp = {"class": "H5T_INTEGER"}
            signed = canonical[1].upper()
            htyp["base"] = f"H5T_STD_{signed}{8 * bytelen}{order}"
        else:
            raise NotImplementedError()
        logging.info("%s %s", typ, htyp)
        logging.info("%s %s, %s", typ, dtype, canonical)



import pickle
from types import UnionType

import numpy as np
import zmq

from dranspose.data.xspress3 import (
    XspressPacket,
    XspressImage,
)
from dranspose.event import StreamData


def parse(
    data: StreamData,
) -> UnionType:
    """
    Parses a Xspress3 packet, which either gives a start/end message or a tuple with a spectra array

    Arguments:
        data: a frame comming from the Xspress3 tango device

    Returns:
        an XspressPacket
    """
    assert data.typ == "xspress", "wrong packet typ"
    assert data.length >= 1, "wrong number of multiparts"
    headerframe = data.frames[0]
    if isinstance(headerframe, zmq.Frame):
        headerframe = headerframe.bytes
    packet = XspressPacket.validate_json(headerframe)
    print("packets", packet)
    if isinstance(packet, XspressImage):
        assert data.length == 3, "images does not contain data"
        bufframe = data.frames[1]
        if isinstance(bufframe, zmq.Frame):
            bufframe = bufframe.bytes
        buf = np.frombuffer(bufframe, dtype=packet.type)
        img = buf.reshape(packet.shape)
        metaframe = data.frames[2]
        if isinstance(metaframe, zmq.Frame):
            metaframe = metaframe.bytes
        meta = pickle.loads(metaframe)
        # meta description: ocr[0], AllEvents[0], AllGood[0], ClockTicks[0],
        #                   TotalTicks[0], ResetTicks[0], event_widths, dtc[0]]
        meta = {
            k: v
            for k, v in zip(
                [
                    "ocr",
                    "AllEvents",
                    "AllGood",
                    "ClockTicks",
                    "TotalTicks",
                    "ResetTicks",
                    "event_widths",
                    "dtc",
                ],
                meta,
            )
        }
        packet.data = img
        packet.meta = meta

    return packet



import logging
from datetime import datetime, timezone
from typing import Any

from dranspose.event import ResultData
from dranspose.protocol import ReducerState, StreamName


class TimingReducer:
    def __init__(self, state: ReducerState | None = None, **kwargs: Any) -> None:
        self.publish: dict[StreamName, list[tuple[float, float]]] = {}
        pass

    def process_result(self, result: ResultData, parameters=None):
        print(result)
        logging.debug("result is %s", result.payload)
        if result.payload:
            now = datetime.now(timezone.utc)
            for stream in result.payload:
                if stream not in self.publish:
                    self.publish[stream] = []
                old, delta = result.payload[stream]
                total = (now - old).total_seconds()
                self.publish[stream].append((delta, total))
                logging.debug("result total delta was %s, workerdelta %s", total, delta)

    def finish(self, parameters=None):
        print("finished dummy reducer work")



import argparse
import asyncio
import json
import logging
import os
import signal
import threading
from asyncio import Task
from typing import Literal, Any, Optional

import uvicorn
from pydantic_core import Url
from rlh import RedisStreamLogHandler
import redis as redis

from dranspose.controller import app
from dranspose.distributed import DistributedSettings
from dranspose.helpers import utils
from dranspose.helpers.utils import done_callback
from dranspose.reducer import app as reducer_app
from dranspose.debug_worker import app as debugworker_app
from dranspose.ingesters import *  # noqa: F403, F401
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.protocol import StreamName, WorkerName
from dranspose.worker import Worker, WorkerSettings

from dranspose.replay import replay as run_replay


class CliSettings(DistributedSettings):
    log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR"] = "INFO"


settings = CliSettings()

logging.basicConfig(level=settings.log_level.upper())

logger = logging.getLogger(__name__)


async def main() -> None:
    ins = []
    ins.append(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )
    wos = [
        Worker(WorkerSettings(worker_name=WorkerName("worker" + str(i))))
        for i in range(1, 3)
    ]

    for i in ins + wos:
        t = asyncio.create_task(i.run())
        t.add_done_callback(done_callback)

    rconfig = uvicorn.Config(reducer_app, port=5001, log_level="info")
    rserver = uvicorn.Server(rconfig)
    tr = asyncio.create_task(rserver.serve())
    tr.add_done_callback(done_callback)

    config = uvicorn.Config(app, port=5000, log_level="info")
    server = uvicorn.Server(config)
    server_task = asyncio.create_task(server.serve())
    await server_task

    for i in ins + wos:
        await i.close()


def controller(args: argparse.Namespace) -> None:
    try:
        config = uvicorn.Config(app, port=5000, host=args.host, log_level="info")
        server = uvicorn.Server(config)
        server.run()
    except KeyboardInterrupt:
        print("exiting")


worker_task: Optional[Task[None]] = None


def worker(args: argparse.Namespace) -> None:
    def stop(*args: Any) -> None:
        global worker_task
        if worker_task:
            worker_task.cancel()

    async def run() -> None:
        global worker_task
        loop = asyncio.get_event_loop()
        loop.add_signal_handler(signal.SIGTERM, stop)

        if args.name is None:
            settings = WorkerSettings(worker_class=args.workerclass)
        else:
            settings = WorkerSettings(
                worker_name=args.name, worker_class=args.workerclass
            )

        w = Worker(settings)
        worker_task = asyncio.create_task(w.run())
        await worker_task
        await w.close()

    asyncio.run(run())


ingester_task = None


def ingester(args: argparse.Namespace) -> None:
    print(args.ingesterclass)
    try:
        ing = globals()[args.ingesterclass]
        sett = globals()[args.ingesterclass.replace("Ingester", "Settings")]
    except KeyError:
        ing = utils.import_class(args.ingesterclass)
        parent_name = ing.__bases__[0].__name__
        sett = globals()[parent_name.replace("Ingester", "Settings")]
        logger.info("custom worker class %s", ing)

    def stop(*args: Any) -> None:
        global ingester_task
        if ingester_task:
            ingester_task.cancel()

    async def run() -> None:
        global ingester_task
        loop = asyncio.get_event_loop()
        loop.add_signal_handler(signal.SIGTERM, stop)

        i = ing(
            settings=sett(upstream_url=args.upstream_url, ingester_streams=[args.name]),
        )
        ingester_task = asyncio.create_task(i.run())
        await ingester_task
        await i.close()

    asyncio.run(run())


def reducer(args: argparse.Namespace) -> None:
    try:
        if args.reducerclass:
            os.environ["REDUCER_CLASS"] = args.reducerclass
        config = uvicorn.Config(
            reducer_app, port=args.port or 5000, host=args.host, log_level="info"
        )
        server = uvicorn.Server(config)
        server.run()
    except KeyboardInterrupt:
        print("exiting")


def http_ingester(args: argparse.Namespace) -> None:
    try:
        os.environ["INGESTER_STREAMS"] = json.dumps([args.name])
        config = uvicorn.Config(
            args.ingesterapp, port=args.port or 5000, host=args.host, log_level="info"
        )
        server = uvicorn.Server(config)
        server.run()
    except KeyboardInterrupt:
        print("exiting")


def debugworker(args: argparse.Namespace) -> None:
    try:
        if args.name:
            os.environ["WORKER_NAME"] = args.name
        config = uvicorn.Config(
            debugworker_app, port=5000, host=args.host, log_level="info"
        )
        server = uvicorn.Server(config)
        server.run()
    except KeyboardInterrupt:
        print("exiting")


def combined(args: argparse.Namespace) -> None:
    asyncio.run(main())


def replay(args: argparse.Namespace) -> None:
    keepalive = None
    if args.keep_alive:
        keepalive = threading.Event()
    run_replay(
        args.workerclass,
        args.reducerclass,
        args.files,
        args.source,
        args.parameters,
        args.port,
        keepalive,
        args.nworkers,
    )


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="dranspose", description="Transposes Streams")

    subparsers = parser.add_subparsers(title="commands", dest="subcommand")

    parser_ctrl = subparsers.add_parser("controller", help="run controller")
    parser_ctrl.set_defaults(func=controller)
    parser_ctrl.add_argument("--host", help="host to listen on")

    parser_reducer = subparsers.add_parser("reducer", help="run reducer")
    parser_reducer.set_defaults(func=reducer)
    parser_reducer.add_argument("--host", help="host to listen on")
    parser_reducer.add_argument("-p", "--port", help="port to listen on", type=int)
    parser_reducer.add_argument(
        "-c",
        "--reducerclass",
        help="reducer class e.g. 'src.reducer:FluorescenceReducer'",
    )

    parser_debugworker = subparsers.add_parser("debugworker", help="run debugworker")
    parser_debugworker.set_defaults(func=debugworker)
    parser_debugworker.add_argument("--host", help="host to listen on")
    parser_debugworker.add_argument(
        "-n",
        "--name",
        help="debug worker name",
    )

    parser_worker = subparsers.add_parser("worker", help="run worker")
    parser_worker.set_defaults(func=worker)
    parser_worker.add_argument("-n", "--name", help="worker name (must not contain :)")
    parser_worker.add_argument(
        "-c", "--workerclass", help="worker class e.g. 'src.worker:FluorescenceWorker'"
    )

    parser_ingester = subparsers.add_parser("ingester", help="run ingester")
    parser_ingester.set_defaults(func=ingester)
    parser_ingester.add_argument(
        "-c", "--ingesterclass", help="Ingester Class", required=True
    )
    parser_ingester.add_argument(
        "-u", "--upstream_url", help="Where to connect to upstream", required=False
    )
    parser_ingester.add_argument(
        "-n", "--name", help="Name of the ingester", required=True
    )

    parser_http_ingester = subparsers.add_parser(
        "http_ingester", help="run http ingester"
    )
    parser_http_ingester.set_defaults(func=http_ingester)
    parser_http_ingester.add_argument("--host", help="host to listen on")
    parser_http_ingester.add_argument(
        "-c", "--ingesterapp", help="Ingester App", required=True
    )
    parser_http_ingester.add_argument("-p", "--port", help="Where to listen on")
    parser_http_ingester.add_argument(
        "-n", "--name", help="Name of the ingester", required=True
    )

    parser_all = subparsers.add_parser(
        "combined", help="run all parts in a single process"
    )
    parser_all.set_defaults(func=combined)

    parser_replay = subparsers.add_parser(
        "replay", help="run replay of ingester recorded files"
    )
    parser_replay.set_defaults(func=replay)
    parser_replay.add_argument(
        "-w",
        "--workerclass",
        help="worker class e.g. 'src.worker:FluorescenceWorker'",
        required=True,
    )
    parser_replay.add_argument(
        "-r",
        "--reducerclass",
        help="reducer class e.g. 'src.reducer:FluorescenceReducer'",
        required=True,
    )
    parser_replay.add_argument(
        "-f", "--files", nargs="+", help="List of files to replay"
    )
    parser_replay.add_argument(
        "-s", "--source", help="python class with sources to get data from"
    )
    parser_replay.add_argument(
        "-p", "--parameters", help="parameter file, json or pickle"
    )
    parser_replay.add_argument(
        "--port", type=int, help="expose the reducer result via http on port"
    )
    parser_replay.add_argument(
        "--keep-alive", action="store_true", help="keep the server alive after finish"
    )
    parser_replay.add_argument(
        "--nworkers",
        help="number of workers to distribute events to",
        default=2,
        type=int,
    )

    return parser


def run() -> None:
    parser = create_parser()
    args: argparse.Namespace = parser.parse_args()

    # Check if a subcommand is provided
    if not getattr(args, "subcommand", None):
        parser.print_help()
    else:
        if args.subcommand != "replay":
            logging_redis = redis.from_url(
                f"{settings.redis_dsn}?decode_responses=True&protocol=3"
            )

            root_logger = logging.getLogger()
            handler = RedisStreamLogHandler(
                redis_client=logging_redis,
                stream_name="dranspose_logs",
                maxlen=1000,
                fields=["msg", "lineno", "name", "created", "levelname"],
            )
            root_logger.addHandler(handler)
        args.func(args)


if __name__ == "__main__":
    run()



import asyncio
from typing import Awaitable, Callable, Any, Coroutine, Optional

import aiohttp

import pytest
import zmq.asyncio
import zmq
from fastapi import FastAPI
from pydantic import HttpUrl

from dranspose.ingester import Ingester
from dranspose.protocol import (
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
    StreamName,
)

from dranspose.worker import Worker

from dranspose.ingesters.http_sardana import app as custom_app
from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.asyncio
async def test_sardana(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    http_ingester: Callable[[FastAPI, int, dict[str, Any]], Awaitable[None]],
    stream_sardana: Callable[[HttpUrl, int], Coroutine[Any, Any, None]],
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"))
    await http_ingester(custom_app, 5002, {"ingester_streams": ["sardana"]})

    await wait_for_controller(streams={StreamName("sardana")})
    async with aiohttp.ClientSession() as session:
        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "sardana": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200

    context = zmq.asyncio.Context()

    asyncio.create_task(stream_sardana(HttpUrl("http://localhost:5002"), ntrig - 1))

    content = await wait_for_finish()

    context.destroy()

    assert content == {
        "last_assigned": ntrig + 1,
        "completed_events": ntrig + 1,
        "total_events": ntrig + 1,
        "finished": True,
    }



import asyncio
import gzip
import json
import logging
import multiprocessing
import os
import pickle
import random
import struct
import time
from asyncio import StreamReader, StreamWriter, Task
from dataclasses import dataclass
from datetime import datetime, timezone
from json import JSONDecodeError
from multiprocessing import Process
from typing import (
    Coroutine,
    AsyncIterator,
    Callable,
    Any,
    Optional,
)

import aiohttp
import cbor2
import numpy as np
import pytest
import pytest_asyncio
import uvicorn
import zmq
from _pytest.fixtures import FixtureRequest
from _pytest.logging import LogCaptureFixture
from aiohttp import ClientConnectionError
from fastapi import FastAPI
from pydantic import HttpUrl
from pydantic_core import Url

from dranspose.controller import app
from dranspose.distributed import DistributedService
from dranspose.helpers.utils import cancel_and_wait
from dranspose.ingester import Ingester, IngesterSettings
from dranspose.ingesters.zmqpull_single import ZmqPullSingleIngester
from dranspose.protocol import WorkerName
from dranspose.worker import Worker, WorkerSettings
from dranspose.reducer import app as reducer_app
from dranspose.debug_worker import app as debugworker_app
from dranspose.workload_generator import app as generator_app
from tests.stream1 import AcquisitionSocket

import redis.asyncio as redis

from pytest import Parser

os.environ["BUILD_META_FILE"] = "tests/data/build_git_meta.json"


def pytest_addoption(parser: Parser) -> None:
    parser.addoption(
        "--rust",
        action="store_true",
        dest="rust",
        default=False,
        help="enable rust decorated tests",
    )
    parser.addoption(
        "--k8s",
        action="store_true",
        dest="k8s",
        default=False,
        help="enable k8s remote tests",
    )
    parser.addoption(
        "--observe",
        action="store_true",
        dest="observe",
        default=False,
        help="enable sampling redis content",
    )


class ErrorLoggingHandler(logging.Handler):
    def __init__(self) -> None:
        super().__init__()
        self.error_found = False

    def emit(self, record: logging.LogRecord) -> None:
        # Check if the log message starts with "ERROR"
        if record.levelname == "ERROR":
            if hasattr(record, "message") and (
                "Unclosed connection" in record.message
            ):  # aiohttp benign error
                return
            self.error_found = True


@pytest_asyncio.fixture(autouse=True)
async def fail_on_error_log(caplog: LogCaptureFixture) -> AsyncIterator[None]:
    handler = ErrorLoggingHandler()
    logging.getLogger().addHandler(handler)

    yield

    # After the test runs, check if any "ERROR" log was encountered
    logging.getLogger().removeHandler(handler)

    if handler.error_found:
        pytest.fail("Test failed due to log message starting with 'ERROR'")


@pytest_asyncio.fixture()
async def controller() -> AsyncIterator[None]:
    config = uvicorn.Config(app, port=5000, log_level="debug")
    server = uvicorn.Server(config)
    server_task = asyncio.create_task(server.serve())
    while server.started is False:
        await asyncio.sleep(0.1)
    yield
    server.should_exit = True
    await server_task
    await asyncio.sleep(0.1)


@dataclass
class AsyncDistributed:
    instance: DistributedService
    task: Task[Any]

    async def stop(self) -> None:
        await self.instance.close()
        await cancel_and_wait(self.task)


@dataclass
class ProcessDistributed:
    instance: DistributedService
    process: multiprocessing.Process
    queue: Any

    async def stop(self) -> None:
        self.queue.put("stop")
        self.process.join()


@dataclass
class ExternalDistributed:
    instance: DistributedService
    process: asyncio.subprocess.Process
    log_tasks: list[Task[Any]]

    async def stop(self) -> None:
        logging.warning("stopping process")
        self.process.terminate()
        await self.process.wait()
        logging.warning("stopping log task")
        for t in self.log_tasks:
            t.cancel()


@pytest_asyncio.fixture
async def create_worker() -> AsyncIterator[
    Callable[[WorkerName], Coroutine[None, None, Worker]]
]:
    workers: list[AsyncDistributed | ProcessDistributed] = []

    async def _make_worker(
        name: WorkerName | Worker, subprocess: bool = False
    ) -> Worker:
        if not isinstance(name, Worker):
            worker = Worker(WorkerSettings(worker_name=name))
        else:
            worker = name

        if subprocess:
            q: Any = multiprocessing.Queue()
            p = Process(target=worker.sync_run, args=(q,), daemon=True)
            p.start()
            workers.append(ProcessDistributed(instance=worker, process=p, queue=q))
        else:
            worker_task = asyncio.create_task(worker.run())
            workers.append(AsyncDistributed(instance=worker, task=worker_task))

        return worker

    yield _make_worker

    for wo in workers:
        await wo.stop()


@pytest_asyncio.fixture
async def create_ingester(
    request: FixtureRequest,
) -> AsyncIterator[Callable[[Ingester], Coroutine[None, None, Ingester]]]:
    ingesters: list[AsyncDistributed | ProcessDistributed | ExternalDistributed] = []

    async def forward_pipe(
        out: StreamReader | None, settings: IngesterSettings
    ) -> None:
        if out is None:
            return
        while True:
            try:
                data = await out.readline()
                line = data.decode("ascii").rstrip()
                if len(line) > 0:
                    fn = logging.error
                    parts = line.split("]")
                    if len(parts) > 1:
                        if "INFO" in parts[0]:
                            fn = logging.info
                        elif "DEBUG" in parts[0]:
                            fn = logging.debug
                        # line = "]".join(parts[1:])
                    fn("rust_%s: %s", settings.ingester_streams[0], line)
                else:
                    break
            except Exception as e:
                logging.error("outputing broke %s", e.__repr__())
                break

    async def _make_ingester(inst: Ingester, subprocess: bool = False) -> Ingester:
        if request.config.getoption("rust"):
            logging.warning("replace ingester with rust")
            if isinstance(inst, ZmqPullSingleIngester):
                logging.warning("ues settings %s", inst._streaming_single_settings)
                proc = await asyncio.create_subprocess_exec(
                    "./perf/target/debug/fast_ingester",
                    "--stream",
                    inst._streaming_single_settings.ingester_streams[0],
                    "--upstream-url",
                    str(inst._streaming_single_settings.upstream_url),
                    "--ingester-url",
                    str(inst._streaming_single_settings.ingester_url),
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE,
                )
                output = asyncio.create_task(
                    forward_pipe(proc.stdout, inst._streaming_single_settings)
                )
                outerr = asyncio.create_task(
                    forward_pipe(proc.stderr, inst._streaming_single_settings)
                )
                ingesters.append(
                    ExternalDistributed(
                        instance=inst, process=proc, log_tasks=[output, outerr]
                    )
                )
                return inst
        if subprocess:
            q: Any = multiprocessing.Queue()
            p = Process(target=inst.sync_run, args=(q,), daemon=True)
            p.start()
            ingesters.append(ProcessDistributed(instance=inst, process=p, queue=q))
        else:
            ingester_task = asyncio.create_task(inst.run())
            ingesters.append(AsyncDistributed(instance=inst, task=ingester_task))
        return inst

    yield _make_ingester

    for ing in ingesters:
        await ing.stop()


@pytest_asyncio.fixture()
async def reducer(
    tmp_path: Any,
) -> AsyncIterator[Callable[[Optional[str]], Coroutine[None, None, None]]]:
    server_tasks = []

    async def start_reducer(custom: Optional[str] = None) -> None:
        envfile = None
        if custom:
            p = tmp_path / "reducer.env"
            p.write_text(
                f"""
                REDUCER_CLASS="{custom}"
                """
            )
            envfile = str(p)
        config = uvicorn.Config(
            reducer_app, port=5001, log_level="debug", env_file=envfile
        )
        server = uvicorn.Server(config)
        server_tasks.append((server, asyncio.create_task(server.serve())))
        while server.started is False:
            await asyncio.sleep(0.1)
        if custom:
            del os.environ["REDUCER_CLASS"]

    yield start_reducer

    for server, task in server_tasks:
        server.should_exit = True
        await task

    await asyncio.sleep(0.1)


@pytest_asyncio.fixture()
async def http_ingester(
    tmp_path: Any,
) -> AsyncIterator[
    Callable[[FastAPI, int, dict[str, Any]], Coroutine[None, None, None]]
]:
    server_tasks = []

    async def start_ingester(
        custom_app: FastAPI, port: int = 5002, env: Optional[dict[str, Any]] = None
    ) -> None:
        envfile = None
        if env:
            p = tmp_path / "http_ingester.env"
            p.write_text(
                "\n".join(
                    [
                        f"""
                {var.upper()}='{json.dumps(val)}'
                """
                        for var, val in env.items()
                    ]
                )
            )
            envfile = str(p)
        config = uvicorn.Config(
            custom_app, port=port, log_level="debug", env_file=envfile
        )
        server = uvicorn.Server(config)
        server_tasks.append((server, asyncio.create_task(server.serve())))
        while server.started is False:
            await asyncio.sleep(0.1)

    yield start_ingester

    for server, task in server_tasks:
        server.should_exit = True
        await task

    await asyncio.sleep(0.1)


@pytest_asyncio.fixture()
async def debug_worker(
    tmp_path: Any,
) -> AsyncIterator[
    Callable[[Optional[str], Optional[list[str]]], Coroutine[None, None, None]]
]:
    server_tasks = []

    async def start_debug_worker(
        name: Optional[str] = None, tags: Optional[list[str]] = None
    ) -> None:
        envfile = None
        if tags:
            p = tmp_path / "debugworker.env"
            text = f"""
                WORKER_TAGS='{json.dumps(tags)}'
                """
            if name:
                text += f"""
                    WORKER_NAME = '{name}'
                    """

            p.write_text(text)
            envfile = str(p)
        config = uvicorn.Config(
            debugworker_app, port=5002, log_level="debug", env_file=envfile
        )
        server = uvicorn.Server(config)
        server_tasks.append((server, asyncio.create_task(server.serve())))
        while server.started is False:
            await asyncio.sleep(0.1)
        if tags:  # turns out uvicorn just dumps the envfile in the process env
            del os.environ["WORKER_TAGS"]
            if name:
                del os.environ["WORKER_NAME"]

    yield start_debug_worker

    for server, task in server_tasks:
        server.should_exit = True
        await task

    await asyncio.sleep(0.1)


class UvicornServer(multiprocessing.Process):
    def __init__(self, config: uvicorn.Config):
        super().__init__()
        self.server = uvicorn.Server(config=config)
        self.config = config
        logging.info("started server with config %s", config)

    def stop(self) -> None:
        self.terminate()

    def run(self, *args: Any, **kwargs: Any) -> None:
        self.server.run()


@pytest_asyncio.fixture()
async def workload_generator(
    tmp_path: Any,
) -> AsyncIterator[Callable[[int], Coroutine[None, None, None]]]:
    server_tasks = []

    async def start_generator(port: int = 5003) -> None:
        config = uvicorn.Config(generator_app, port=port, log_level="debug")
        instance = UvicornServer(config=config)
        instance.start()

        async with aiohttp.ClientSession() as session:
            while True:
                try:
                    st = await session.get(f"http://localhost:{port}/api/v1/finished")
                    if st.status == 200:
                        break
                    else:
                        await asyncio.sleep(0.5)
                except ClientConnectionError:
                    await asyncio.sleep(0.5)

        server_tasks.append(instance)

    yield start_generator

    for server in server_tasks:
        server.stop()

    await asyncio.sleep(0.1)


@pytest_asyncio.fixture
async def stream_eiger() -> Callable[
    [zmq.Context[Any], int, int], Coroutine[Any, Any, None]
]:
    async def _make_eiger(
        ctx: zmq.Context[Any], port: int, nframes: int, frame_time: float = 0.1
    ) -> None:
        socket = AcquisitionSocket(ctx, Url(f"tcp://*:{port}"))
        acq = await socket.start(filename="")
        width = 1475
        height = 831
        for frameno in range(nframes):
            img = np.zeros((width, height), dtype=np.uint16)
            for _ in range(20):
                img[random.randint(0, width - 1)][
                    random.randint(0, height - 1)
                ] = random.randint(0, 10)
            extra = {"timestamps": {"dummy": datetime.now(timezone.utc).isoformat()}}
            await acq.image(img, img.shape, frameno, extra_fields=extra)
            await asyncio.sleep(frame_time)
        await acq.close()
        await socket.close()

    return _make_eiger


@pytest_asyncio.fixture
async def stream_pkls() -> Callable[
    [
        zmq.Context[Any],
        int,
        os.PathLike[Any] | str,
        float,
        int,
        Optional[int],
        Optional[int],
    ],
    Coroutine[Any, Any, None],
]:
    async def _make_pkls(
        ctx: zmq.Context[Any],
        port: int,
        filename: os.PathLike[Any] | str,
        frame_time: float = 0.1,
        typ: int = zmq.PUSH,
        begin: Optional[int] = None,
        end: Optional[int] = None,
    ) -> None:
        socket: zmq.Socket[Any] = ctx.socket(typ)
        socket.bind(f"tcp://*:{port}")
        if begin is None and end is None:
            for _ in range(3):
                await socket.send_multipart([b"emptyness"])
                await asyncio.sleep(0.1)
        base, ext = os.path.splitext(filename)
        if ext == ".gz":
            f = gzip.open(filename, "rb")
            _, ext = os.path.splitext(base)
        else:
            f = open(filename, "rb")
        i = 0
        while True:
            try:
                if "cbor" in ext:
                    frames = cbor2.load(f)
                else:
                    frames = pickle.load(f)
                send = True
                if i < (begin or 0):
                    send = False
                if end:
                    if i >= end:
                        send = False
                if send:
                    await socket.send_multipart(frames)
                    await asyncio.sleep(frame_time)
            except EOFError:
                break
        f.close()
        socket.close()

    return _make_pkls


@pytest_asyncio.fixture
async def stream_orca() -> Callable[
    [zmq.Context[Any], int, int], Coroutine[Any, Any, None]
]:
    async def _make_orca(ctx: zmq.Context[Any], port: int, nframes: int) -> None:
        socket = AcquisitionSocket(ctx, Url(f"tcp://*:{port}"))
        acq = await socket.start(filename="")
        width = 2000
        height = 4000
        for frameno in range(nframes):
            img = np.zeros((width, height), dtype=np.uint16)
            for _ in range(20):
                img[random.randint(0, width - 1)][
                    random.randint(0, height - 1)
                ] = random.randint(0, 10)
            await acq.image(img, img.shape, frameno)
            await asyncio.sleep(0.1)
        await acq.close()
        await socket.close()

    return _make_orca


@pytest_asyncio.fixture
async def stream_small() -> Callable[
    [zmq.Context[Any], int | str, int], Coroutine[Any, Any, None]
]:
    async def _make_alba(
        ctx: zmq.Context[Any], port: int | str, nframes: int, frame_time: float = 0.1
    ) -> None:
        if type(port) is str:
            socket = AcquisitionSocket(ctx, Url(port))
        else:
            socket = AcquisitionSocket(ctx, Url(f"tcp://*:{port}"))
        acq = await socket.start(filename="")
        val = np.zeros((10000,), dtype=np.float64)
        start = time.perf_counter()
        for frameno in range(nframes):
            await acq.image(val, val.shape, frameno)
            if frame_time:
                await asyncio.sleep(frame_time)
            if frameno % 1000 == 0:
                end = time.perf_counter()
                logging.info(
                    "send 1000 packets took %s: %lf p/s",
                    end - start,
                    1000 / (end - start),
                )
                start = end
        await acq.close()
        await socket.close()

    return _make_alba


@pytest_asyncio.fixture
async def stream_small_xrd() -> Callable[
    [zmq.Context[Any], int | str, int], Coroutine[Any, Any, None]
]:
    async def _make_xrd(
        ctx: zmq.Context[Any], port: int | str, nframes: int, frame_time: float = 0.1
    ) -> None:
        if type(port) is str:
            socket = AcquisitionSocket(ctx, Url(port))
        else:
            socket = AcquisitionSocket(ctx, Url(f"tcp://*:{port}"))
        acq = await socket.start(filename="output/xrd.h5")
        start = time.perf_counter()
        for frameno in range(nframes):
            base = np.zeros((10, 10), dtype=np.uint16)
            for x in range(base.shape[0]):
                for y in range(base.shape[1]):
                    base[x, y] = max(
                        np.sin(
                            (
                                (x - base.shape[0] / 2) ** 2
                                + (y - base.shape[1] / 2) ** 2
                            )
                            / (3 + frameno)
                        )
                        * 10
                        + 10
                        + random.randrange(-2, 2),
                        0,
                    )
            await acq.image(base, base.shape, frameno)
            if frame_time:
                await asyncio.sleep(frame_time)
            if frameno % 1000 == 0:
                end = time.perf_counter()
                logging.info(
                    "send 1000 packets took %s: %lf p/s",
                    end - start,
                    1000 / (end - start),
                )
                start = end
        await acq.close()
        await socket.close()

    return _make_xrd


@pytest_asyncio.fixture
async def time_beacon() -> Callable[
    [zmq.Context[Any], int, int], Coroutine[Any, Any, None]
]:
    async def _make_time(
        ctx: zmq.Context[Any],
        port: int,
        nframes: int,
        frame_time: float = 0.1,
        flags: int = 0,
    ) -> None:
        sout = ctx.socket(zmq.PUSH)
        sout.bind(f"tcp://*:{port}")
        start = time.perf_counter()
        for i in range(nframes):
            data = struct.pack(">d", time.perf_counter())
            try:
                await sout.send(data, flags=flags)
            except zmq.Again:
                pass
            if frame_time is not None:
                await asyncio.sleep(frame_time)
            if i % 1000 == 0:
                end = time.perf_counter()
                logging.info(
                    "send 1000 time packets took %s: %lf p/s",
                    end - start,
                    1000 / (end - start),
                )
                start = end
        sout.close(linger=0)

    return _make_time


@pytest_asyncio.fixture
async def stream_sardana() -> Callable[[HttpUrl, int], Coroutine[Any, Any, None]]:
    async def _make_sardana(url: HttpUrl, nframes: int) -> None:
        async with aiohttp.ClientSession() as session:
            descurl = HttpUrl.build(
                scheme=url.scheme,
                username=url.username,
                password=url.password,
                host=url.host or "localhost",
                port=url.port,
                path="v1/data_desc",
            )
            recurl = HttpUrl.build(
                scheme=url.scheme,
                username=url.username,
                password=url.password,
                host=url.host or "localhost",
                port=url.port,
                path="v1/record_data",
            )
            endurl = HttpUrl.build(
                scheme=url.scheme,
                username=url.username,
                password=url.password,
                host=url.host or "localhost",
                port=url.port,
                path="v1/record_end",
            )
            st = await session.post(
                str(descurl),
                json={
                    "column_desc": [
                        {
                            "name": "point_nb",
                            "label": "#Pt No",
                            "dtype": "int64",
                            "shape": [],
                        },
                        {
                            "min_value": 0,
                            "max_value": 0.1,
                            "instrument": "",
                            "name": "dummy_mot_01",
                            "label": "dummy_mot_01",
                            "dtype": "float64",
                            "shape": [],
                            "is_reference": True,
                        },
                        {
                            "name": "timestamp",
                            "label": "dt",
                            "dtype": "float64",
                            "shape": [],
                        },
                    ],
                    "ref_moveables": ["dummy_mot_01"],
                    "estimatedtime": -2.6585786096205566,
                    "total_scan_intervals": 2,
                    "starttime": "Mon Apr 17 14:19:35 2023",
                    "title": "burstscan dummy_mot_01 0.0 0.1 2 50",
                    "counters": [
                        "tango://b-v-femtomax-csdb-0.maxiv.lu.se:10000/expchan/oscc_02_seq_ctrl/1",
                        "tango://b-v-femtomax-csdb-0.maxiv.lu.se:10000/expchan/panda_femtopcap_ctrl/4",
                        "tango://b-v-femtomax-csdb-0.maxiv.lu.se:10000/expchan/panda_femtopcap_ctrl/5",
                    ],
                    "scanfile": ["stream.daq", "tests_03.h5"],
                    "scandir": "/data/staff/femtomax/20230413",
                    "serialno": 20338,
                },
            )
            assert st.status == 200

            for frameno in range(nframes):
                st = await session.post(
                    str(recurl),
                    json={
                        "point_nb": frameno,
                        "dummy_mot_01": 0,
                        "timestamp": 0.6104741096496582,
                    },
                )
                assert st.status == 200
                await asyncio.sleep(0.1)
            st = await session.post(str(endurl), json={})
            assert st.status == 200

    return _make_sardana


@pytest_asyncio.fixture
async def stream_pcap() -> AsyncIterator[
    Callable[[int, int, int], Coroutine[None, None, None]]
]:
    server_tasks = []

    async def _make_pcap(
        nframes: int = 10, port: int = 8889, repetitions: int = 1
    ) -> None:
        async def handle_client(reader: StreamReader, writer: StreamWriter) -> None:
            request = (await reader.read(255)).decode("utf8")
            if request.strip() != "":
                logging.error("bad initial hello %s", request)
            writer.write(b"OK\n")
            await writer.drain()
            for i in range(repetitions):
                header = b"""arm_time: 2023-12-19T07:51:12.754Z
missed: 0
process: Scaled
format: ASCII
fields:
 PCAP.BITS0 uint32 Value
 INENC1.VAL double Min scale: 1 offset: 0 units:
 INENC1.VAL double Max scale: 1 offset: 0 units:
 PCAP.TS_TRIG double Value scale: 8e-09 offset: 0 units: s
 INENC1.VAL double Mean scale: 1 offset: 0 units:
 SFP3_SYNC_IN.POS1 double Mean scale: 1 offset: 0 units: (null)

"""
                writer.write(header)
                await writer.drain()

                data = b""" 301989857 0 0 0.32175968 0 592910924.7
 33554401 0 0 0.43175968 0 592910919
 33554401 0 0 0.54175968 0 592910930.9
 33554401 0 0 0.65175968 0 592910930.2
 301989857 0 0 0.76175968 0 592910946.1
 301989857 0 0 0.87175968 0 592910969.6
 301989857 0 0 0.98175968 0 592911000.4
 301989857 0 0 1.09175968 0 592911008.9
 301989857 0 0 1.20175968 0 592910962.2
 301989857 0 0 1.31175968 0 592910916.7
 301989857 0 0 1.42175968 0 592910829.8""".split(
                    b"\n"
                )

                for i in range(nframes):
                    writer.write(data[i % 11] + b"\n")
                    await writer.drain()
                    await asyncio.sleep(0.1)

                writer.write(f"END {nframes} Disarmed\n".encode("utf8"))
                await writer.drain()
            writer.close()

        server = await asyncio.start_server(handle_client, "localhost", port)
        server_tasks.append(asyncio.create_task(server.serve_forever()))

        logging.warning("started server %s", server.is_serving())
        # async with server:
        #    await server.serve_forever()

    yield _make_pcap

    for task in server_tasks:
        await cancel_and_wait(task)


multiscan_trace = """

OK
arm_time: 2023-12-19T08:49:51.103Z
missed: 0
process: Scaled
format: ASCII
fields:
 PCAP.BITS0 uint32 Value
 INENC1.VAL double Min scale: 1 offset: 0 units:
 INENC1.VAL double Max scale: 1 offset: 0 units:
 PCAP.TS_TRIG double Value scale: 8e-09 offset: 0 units: s
 INENC1.VAL double Mean scale: 1 offset: 0 units:
 SFP3_SYNC_IN.POS1 double Mean scale: 1 offset: 0 units: (null)

 301989857 0 0 0.290445128 0 592659802.5
 33554401 0 0 0.400445128 0 592659790.2
 301989857 0 0 0.510445128 0 592659777.6
 301989857 0 0 0.620445128 0 592659776.1
 301989857 0 0 0.730445128 0 592659780.1
 301989857 0 0 0.840445128 0 592659786.7
 301989857 0 0 0.950445128 0 592659778.1
 301989857 0 0 1.060445128 0 592659753.3
 33554401 0 0 1.170445128 0 592659740.2
 301989857 0 0 1.280445128 0 592659740.5
 301989857 0 0 1.390445128 0 592659745.6
END 11 Disarmed
arm_time: 2023-12-19T08:49:59.911Z
missed: 0
process: Scaled
format: ASCII
fields:
 PCAP.BITS0 uint32 Value
 INENC1.VAL double Min scale: 1 offset: 0 units:
 INENC1.VAL double Max scale: 1 offset: 0 units:
 PCAP.TS_TRIG double Value scale: 8e-09 offset: 0 units: s
 INENC1.VAL double Mean scale: 1 offset: 0 units:
 SFP3_SYNC_IN.POS1 double Mean scale: 1 offset: 0 units: (null)

 301989857 0 0 0.280505096 0 592659819.6
 301989857 0 0 0.390505096 0 592659807.4
 301989857 0 0 0.500505096 0 592659823.6
 33554401 0 0 0.610505096 0 592659827.9
 33554401 0 0 0.720505096 0 592659835.1
 301989857 0 0 0.830505096 0 592659817.4
 301989857 0 0 0.940505096 0 592659828.3
 301989857 0 0 1.050505096 0 592659823.2
 301989857 0 0 1.160505096 0 592659851.4
 33554401 0 0 1.270505096 0 592659838.7
 33554401 0 0 1.380505096 0 592659842.7
END 11 Disarmed"""


def sample_table_to_md(table: dict[str, Any], filename: str) -> None:
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, "w") as f:
        col_order = sorted(table.keys())  # , key=lambda name: name[::-1])
        logging.info("column order is %s", col_order)
        f.write(f'| {" | ".join(col_order)} |\n')
        f.write(f'| {" | ".join(["---" for _ in table])} |\n')
        last_line: tuple[Any, ...] = tuple([""] * len(table))
        for line in zip(*[table[col] for col in col_order]):
            if line == last_line:
                continue
            change_line = []
            for old, new in zip(last_line, line):
                if new is None:
                    change_line.append("absent")
                elif old == new:
                    change_line.append('--"--')
                else:
                    change = new
                    olddata = None
                    if isinstance(old, str):
                        try:
                            olddata = json.loads(old)
                        except JSONDecodeError:
                            pass
                    if isinstance(new, str):
                        try:
                            data = json.loads(new)
                            diff = data
                            change = ""
                            if olddata is not None:
                                diff = {}
                                for key, item in data.items():
                                    if key in olddata and olddata[key] == item:
                                        continue
                                    diff[key] = item
                                change = "**changed**<br>"
                            change += (
                                json.dumps(diff, indent=2)
                                .replace("\n", "<br>")
                                .replace(" ", "&nbsp;")
                            )
                        except JSONDecodeError:
                            pass
                    elif isinstance(new, bytes):
                        change = str(new[:30]) + f" ({len(new)} bytes)"
                    elif isinstance(new, list):
                        change = ""
                        for entry in new:
                            change += "**" + entry[0] + "**<br>"
                            data = json.loads(entry[1]["data"])
                            pretty = (
                                json.dumps(data, indent=2)
                                .replace("\n", "<br>")
                                .replace(" ", "&nbsp;")
                            )
                            change += pretty + "<br>"
                    change_line.append(change)
            f.write(f'| {" | ".join(map(str, change_line))} |\n')
            last_line = line
    logging.info("written redis observation to %s", filename)


def sample_table_to_html(
    table: dict[str, str | list[tuple[str, str]] | bytes], filename: str
) -> None:
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, "w") as f:
        col_order = sorted(table.keys())  # , key=lambda name: name[::-1])
        logging.info("column order is %s", col_order)
        f.write(
            """
        <html>
        <head>
        <style>
        th {
          position: sticky;
          top: 50px;
          background: white;
        }
        </style>
        </head>
        <body style="padding-top:50px;">
        <table>
        <thead>
          <tr>
        """
        )
        for col in col_order:
            f.write(f"<th>{col}</th>")
        f.write(
            """  </tr>
        </thead>
        <tbody>
        """
        )
        last_line: tuple[Any, ...] = tuple([""] * len(table))
        for line in zip(*[table[col] for col in col_order]):
            if line == last_line:
                continue
            change_line = []
            for old, new in zip(last_line, line):
                if new is None:
                    change_line.append("absent")
                elif old == new:
                    change_line.append('--"--')
                else:
                    change = new
                    olddata = None
                    if isinstance(old, str):
                        try:
                            olddata = json.loads(old)
                        except JSONDecodeError:
                            pass
                    if isinstance(new, str):
                        try:
                            data = json.loads(new)
                            diff = data
                            change = ""
                            if olddata is not None:
                                diff = {}
                                for key, item in data.items():
                                    if key in olddata and olddata[key] == item:
                                        continue
                                    diff[key] = item
                                change = "<b>changed</b><br>"
                            change += "<pre>"
                            change += json.dumps(diff, indent=2)
                            change += "</pre>"
                        except JSONDecodeError:
                            pass
                    elif isinstance(new, bytes):
                        change = str(new[:30]) + f" ({len(new)} bytes)"
                    elif isinstance(new, list):
                        change = ""
                        for entry in new:
                            change += "<b>" + entry[0] + "</b><br><pre>"
                            data = json.loads(entry[1]["data"])
                            pretty = json.dumps(data, indent=2)
                            change += pretty + "</pre><br>"
                    change_line.append(change)
            f.write("<tr>")
            for el in change_line:
                f.write(f"<td>{str(el)}</td>")
            f.write("<tr>")
            last_line = line
        f.write(
            """
        </tbody>
        </table></body></html>"""
        )
    logging.info("written redis observation to %s", filename)


async def sample_redis(filename: str) -> None:
    r = redis.Redis(host="localhost", port=6379, decode_responses=True, protocol=3)
    r_raw = redis.Redis(host="localhost", port=6379, decode_responses=False, protocol=3)
    table: dict[str, Any] = {}
    try:
        tick = 0
        start_ids: dict[str, str] = {}
        while True:
            keys = await r.keys("dranspose:*")
            logging.debug("all redis keys %s", keys)
            for key in keys:
                if key.startswith("dranspose:parameters:"):
                    data = await r_raw.get(key)
                elif any(
                    [
                        match in key
                        for match in [
                            "dranspose:ready:",
                            "dranspose:controller:updates",
                            "dranspose:assigned:",
                        ]
                    ]
                ):
                    start = start_ids.get(key, "-")
                    if start != "-":
                        start = "(" + start
                    data = await r.xrange(key, start, "+")
                    if len(data) > 0:
                        start_ids[key] = data[-1][0]
                    logging.debug("data is %s", data)
                else:
                    data = await r.get(key)
                if key not in table:
                    table[key] = [None] * tick
                table[key].append(data)
                logging.debug("key %s contains %s", key, data)
            for missing in table:
                if missing not in keys:
                    table[missing].append(None)
            await asyncio.sleep(0.1)
            tick += 1
    except asyncio.exceptions.CancelledError:
        # sample_table_to_md(table, filename)
        sample_table_to_html(table, filename)
        await r.aclose()


@pytest_asyncio.fixture(autouse=True)
async def observe_redis(request: FixtureRequest) -> AsyncIterator[None]:
    if request.config.getoption("observe"):
        filename = (
            "redis-observations/"
            + request.node.nodeid.replace("/", ".").replace("::", "--")
            + ".html"
        )
        t = asyncio.create_task(sample_redis(filename))
        yield
        t.cancel()
        await t
    else:
        yield



from typing import Literal

from pydantic import BaseModel, TypeAdapter, ConfigDict


class EigerLegacy(BaseModel):
    pass


class EigerLegacyHeader(EigerLegacy):
    """
    Example:
        ``` py
        EigerLegacyHeader(
            htype='dheader-1.0'
            header_detail='all'
            series=6
            info={'auto_summation': True,
                  'beam_center_x': 0.0,
                  'beam_center_y': 0.0,
                  'bit_depth_image': 32,
                  'bit_depth_readout': 16,
                  'chi_increment': 0.0,
                  'chi_start': 0.0,
                  'compression': 'bslz4',
                  'count_time': 0.0099999,
                  'countrate_correction_applied': True,
                  'countrate_correction_count_cutoff': 83270,
                  'data_collection_date': '2024-03-12T15:29:58.142+01:00',
                  'description': 'Dectris EIGER2 CdTe 9M',
                  'detector_distance': 0.0,
                  'detector_number': 'E-18-0000',
                  'detector_readout_time': 1e-07,
                  'detector_translation': [0.0, 0.0, 0.0],
                  'eiger_fw_version': 'release-2022.1.1',
                  'element': '',
                  'flatfield_correction_applied': True,
                  'frame_count_time': 0.00499996,
                  'frame_period': 0.00499996,
                  'frame_time': 0.01,
                  'kappa_increment': 0.0,
                  'kappa_start': 0.0,
                  'nimages': 3,
                  'ntrigger': 1,
                  'number_of_excluded_pixels': 683998,
                  'omega_increment': 0.0,
                  'omega_start': 0.0,
                  'phi_increment': 0.0,
                  'phi_start': 0.0,
                  'photon_energy': 8041.0,
                  'pixel_mask_applied': True,
                  'roi_mode': '',
                  'sensor_material': 'CdTe',
                  'sensor_thickness': 0.00075,
                  'software_version': '1.8.0',
                  'threshold_energy': 4020.5,
                  'trigger_mode': 'ints',
                  'two_theta_increment': 0.0,
                  'two_theta_start': 0.0,
                  'virtual_pixel_correction_applied': True,
                  'wavelength': 1.5419002416764116,
                  'x_pixel_size': 7.5e-05,
                  'x_pixels_in_detector': 3108,
                  'y_pixel_size': 7.5e-05,
                  'y_pixels_in_detector': 3262}
            appendix={'collect_dict': {
                        'col_id': 128579,
                        'process_dir': '/data/.../process',
                        'target_beam_size_factor': 0,
                        'ssx_mode': 'test',
                        'exp_type': 'test',
                        'shape_id': '2DP1',
                        'col': 0,
                        'row': 0},
                      'dozor_dict': {
                        'x_pixels_in_detector': 3108,
                        'omega_increment': None,
                        'count_time': 0.0099999,
                        'detector_distance': 0.0,
                        'beam_center_y': 0.0,
                        'beam_center_x': 0.0,
                        'y_pixels_in_detector': 3262,
                        'wavelength': 1.541900241676412,
                        'countrate_correction_count_cutoff': 83270,
                        'x_pixel_size': 7.5e-05}
                      }
        )
        ```
    """

    model_config = ConfigDict(extra="allow")

    htype: Literal["dheader-1.0"]


class EigerLegacyImage(EigerLegacy):
    """
    Example:
        ``` py
        EigerLegacyImage(
            htype='dimage-1.0',
            frame=0
            hash='815e4907ce3482ec3cc4d33a1145cd44'
            series=6
            data={'encoding': 'bs32-lz4<',
                  'htype': 'dimage_d-1.0',
                  'shape': [3108, 3262],
                  'size': 754672,
                  'type': 'uint32',
                  'buffer': b'\x00\x00\x00\x00\x02j\xca\xe0...'
            config={'htype': 'dconfig-1.0',
                    'real_time': 9999900,
                    'start_time': 103597359252740,
                    'stop_time': 103597369252640}
        )
        ```
    """

    model_config = ConfigDict(extra="allow")

    htype: Literal["dimage-1.0"]
    frame: int


class EigerLegacyEnd(EigerLegacy):
    """
    Example:
        ``` py
        EigerLegacyEnd(
            htype='dseries_end-1.0'
        )
        ```
    """

    model_config = ConfigDict(extra="allow")

    htype: Literal["dseries_end-1.0"]


EigerLegacyPacket = TypeAdapter(EigerLegacyHeader | EigerLegacyImage | EigerLegacyEnd)
"""
A union type for Eiger Legacy packets
"""



import time
from typing import Any, Iterator, Optional

import zmq
import itertools

from numpy import ndarray

from dranspose.protocol import ZmqUrl


class Acquisition:
    def __init__(self, socket: Any, filename: str, it: Iterator[int]):
        self._socket = socket
        self._filename = filename
        self._msg_number = it

    async def start(self, meta: Any = None) -> None:
        header = {
            "htype": "header",
            "filename": self._filename,
            "msg_number": next(self._msg_number),
        }
        if not meta:
            await self._socket.send_json(header)
        else:
            await self._socket.send_json(header, flags=zmq.SNDMORE)
            await self._socket.send_json(meta)

    async def image(
        self,
        frame: zmq.Frame | ndarray[Any, Any],
        shape: tuple[int, ...],
        frameno: int,
        typ: str = "uint16",
        compression: str = "none",
        extra_fields: Optional[dict[str, Any]] = None,
    ) -> None:
        before = time.perf_counter()
        if extra_fields is None:
            extra_fields = {}
        await self._socket.send_json(
            {
                "htype": "image",
                "frame": frameno,
                "shape": shape,
                "type": typ,
                "compression": compression,
                "msg_number": next(self._msg_number),
                **extra_fields,
            },
            flags=zmq.SNDMORE,
        )
        await self._socket.send(frame, copy=False)
        delay = time.perf_counter() - before
        if delay > 0.05:
            print("sending took", delay)

    async def close(self) -> None:
        await self._socket.send_json(
            {"htype": "series_end", "msg_number": next(self._msg_number)}
        )


class AcquisitionSocket:
    def __init__(
        self, ctx: zmq.Context[Any], bind: ZmqUrl, typ: int = zmq.PUSH
    ) -> None:
        self.data_socket = ctx.socket(typ)
        self.data_socket.bind(str(bind))
        self.msg_number = itertools.count(0)

    async def start(self, filename: str, meta: Any = None) -> Acquisition:
        print("return acq")
        acq = Acquisition(self.data_socket, filename, self.msg_number)
        await acq.start(meta)
        return acq

    async def close(self) -> None:
        self.data_socket.close()



import asyncio
from typing import Awaitable, Callable, Coroutine, Optional

import aiohttp

import pytest
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.tcp_positioncap import TcpPcapIngester, TcpPcapSettings
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)

from dranspose.worker import Worker, WorkerSettings
from tests.utils import wait_for_finish, wait_for_controller


@pytest.mark.asyncio
async def test_panda_pcap(
    stream_pcap: Callable[[int], Coroutine[None, None, None]],
) -> None:
    ntrig = 10
    asyncio.create_task(stream_pcap(ntrig - 1))
    await asyncio.sleep(0.5)
    reader, writer = await asyncio.open_connection("localhost", 8889)
    writer.write(b"\n")
    data = await reader.readline()
    assert data == b"OK\n"
    while data := await reader.readline():
        if data.startswith(b"END"):
            break
    writer.close()
    await writer.wait_closed()

    await asyncio.sleep(1)


@pytest.mark.asyncio
async def test_pcapingester(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[Worker], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_pcap: Callable[[int], Coroutine[None, None, None]],
) -> None:
    await reducer(None)
    await create_worker(
        Worker(
            settings=WorkerSettings(
                worker_name=WorkerName("w1"),
                worker_class="examples.parser.positioncap:PcapWorker",
            ),
        )
    )
    await create_ingester(
        TcpPcapIngester(
            settings=TcpPcapSettings(
                ingester_streams=[StreamName("pcap")],
                upstream_url=Url("tcp://localhost:8889"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("pcap")})
    async with aiohttp.ClientSession() as session:
        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "pcap": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200
        await resp.json()

    asyncio.create_task(stream_pcap(ntrig - 1))

    content = await wait_for_finish()

    print(content)



import json
import logging
import pickle
from typing import Any

from dranspose.data.positioncap import PositionCapValues
from dranspose.event import EventData
from dranspose.middlewares import contrast
from dranspose.middlewares import xspress
from dranspose.middlewares import stream1
from dranspose.middlewares.positioncap import PositioncapParser
from dranspose.parameters import StrParameter, BinaryParameter, ParameterBase
from dranspose.protocol import ParameterName, WorkParameter, StreamName

logger = logging.getLogger(__name__)


class FluorescenceWorker:
    def __init__(self, **kwargs: Any) -> None:
        self.number = 0
        self.pcap = PositioncapParser()

    @staticmethod
    def describe_parameters() -> list[ParameterBase]:
        params = [
            StrParameter(name=ParameterName("roi1"), default="bla"),
            StrParameter(name=ParameterName("file_parameter")),
            BinaryParameter(name=ParameterName("file_parameter_file")),
        ]
        return params

    def process_event(
        self,
        event: EventData,
        parameters: dict[ParameterName, WorkParameter] | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> dict[str, Any] | None:
        logger.debug("using parameters %s", parameters)

        if "dummy" in event.streams:
            logger.info("got dummy data %s", event.streams[StreamName("dummy")])
            temp = json.loads(event.streams[StreamName("dummy")].frames[0].bytes)
            logger.info("temperature is %s", temp)
        if (
            "file_parameter_file" in parameters
            and parameters[ParameterName("file_parameter_file")].value != b""
        ):
            logger.debug(
                "file is given %s",
                parameters[ParameterName("file_parameter_file")].value,
            )
            arr = pickle.loads(parameters[ParameterName("file_parameter_file")].value)
            logger.info("array param %s", arr)

        if "pilatus" in event.streams:
            dat = stream1.parse(event.streams[StreamName("pilatus")])
            logger.info("got pilatus frame %s", dat)

        if "pcap" in event.streams:
            logger.debug("raw pcap:%s", event.streams[StreamName("pcap")].frames[0])
            res = self.pcap.parse(event.streams[StreamName("pcap")])
            if isinstance(res, PositionCapValues):
                logger.info(
                    "got pcap values %s",
                    res,
                )
            else:
                logger.warning("got pcap %s", res)

        if {"contrast", "xspress3"} - set(event.streams.keys()) != set():
            logger.warning(
                "missing streams for this worker, only present %s", event.streams.keys()
            )
            return
        try:
            con = contrast.parse(event.streams[StreamName("contrast")])
        except Exception as e:
            logger.error("failed to parse contrast %s", e.__repr__())
            return

        try:
            spec = xspress.parse(event.streams[StreamName("xspress3")])
        except Exception as e:
            logger.error("failed to parse xspress3 %s", e.__repr__())
            return
        logger.debug("contrast: %s", con)
        logger.debug("spectrum: %s", spec)

        roi_slice = json.loads(parameters[("roi1")].data)

        if con.status == "running":
            # new data
            sx, sy = con.pseudo["x"][0], con.pseudo["y"][0]
            logger.debug("process position %s %s", sx, sy)

            roi1 = spec.data[-1][roi_slice[0] : roi_slice[1]].sum()

            return {"position": (sx, sy), "concentations": {"roi1": roi1}}



import logging

import asyncio
from typing import Awaitable, Callable, Optional, Coroutine, Any
import zmq.asyncio

import aiohttp
import pytest
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.ingesters.tcp_positioncap import (
    TcpPcapIngester,
    TcpPcapSettings,
)

from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)
from dranspose.worker import Worker
from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.asyncio
async def test_multiple_scans(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_pcap: Callable[[int, int, int], Coroutine[None, None, None]],
    stream_eiger: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"))
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )
    await create_ingester(
        TcpPcapIngester(
            settings=TcpPcapSettings(
                ingester_streams=[StreamName("pcap")],
                upstream_url=Url("tcp://localhost:8889"),
                ingester_url=Url("tcp://localhost:10011"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("eiger"), StreamName("pcap")})
    async with aiohttp.ClientSession() as session:
        ntrig = 20
        mp = {
            "eiger": [
                [VirtualWorker(constraint=VirtualConstraint(i)).model_dump(mode="json")]
                if i % 12 < 10
                else None
                for i in range(ntrig)
            ],
            "pcap": [
                [VirtualWorker(constraint=VirtualConstraint(i)).model_dump(mode="json")]
                for i in range(ntrig)
            ],
        }
        logging.info("map is %s", mp)
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json=mp,
        )
        assert resp.status == 200
        await resp.json()

    context = zmq.asyncio.Context()

    asyncio.create_task(stream_eiger(context, 9999, ntrig - 1))

    asyncio.create_task(stream_pcap(9, 8889, 2))

    await wait_for_finish()

    context.destroy()



class FluorescenceReducer:
    @staticmethod
    def describe_parameters():
        return [FileParameter(name="dest_file")]

    def __init__(self, parameters=None):
        self.publish = {"map": {}}

    def process_result(self, result: ResultData, parameters=None):
        if result.payload:
            self.publish["map"][result.payload["position"]] = result.payload["concentations"]



import asyncio
from typing import Awaitable, Callable, Any, Coroutine, Optional

import aiohttp

import pytest
import zmq.asyncio
import zmq
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)

from dranspose.worker import Worker
from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.asyncio
async def test_sequence(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_eiger: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"))
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )
    await wait_for_controller(streams={StreamName("eiger")})
    async with aiohttp.ClientSession() as session:
        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/sequence",
            json={
                "parts": {
                    "main": {
                        "eiger": [
                            [
                                VirtualWorker(
                                    constraint=VirtualConstraint(2 * i)
                                ).model_dump(mode="json")
                            ]
                            for i in range(1, 5)
                        ],
                    },
                    "end": {
                        "eiger": [
                            [
                                VirtualWorker(
                                    constraint=VirtualConstraint(2 * i)
                                ).model_dump(mode="json")
                            ]
                            for i in range(1, 2)
                        ],
                    },
                },
                "sequence": ["main", "main", "end"],
            },
        )
        assert resp.status == 200

    context = zmq.asyncio.Context()

    asyncio.create_task(stream_eiger(context, 9999, ntrig - 1))

    await wait_for_finish()
    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5000/api/v1/mapping")
        content = await st.json()
        assert content == {
            "parts": {
                "main": {
                    "mapping": {
                        "eiger": [
                            [{"tags": ["generic"], "constraint": 2}],
                            [{"tags": ["generic"], "constraint": 4}],
                            [{"tags": ["generic"], "constraint": 6}],
                            [{"tags": ["generic"], "constraint": 8}],
                        ]
                    }
                },
                "end": {
                    "mapping": {"eiger": [[{"tags": ["generic"], "constraint": 2}]]}
                },
                "reserved_start_end": {
                    "mapping": {"eiger": [[{"tags": ["generic"], "constraint": None}]]}
                },
            },
            "sequence": [
                "reserved_start_end",
                "main",
                "main",
                "end",
                "reserved_start_end",
            ],
        }

    context.destroy()



import asyncio
import json
import logging
import pickle
import threading
from typing import Any

import aiohttp
import pytest

from dranspose.replay import replay


@pytest.mark.asyncio
async def test_replay(
    tmp_path: Any,
) -> None:
    par_file = tmp_path / "parameters.json"

    with open(par_file, "w") as f:
        json.dump([{"name": "roi1", "data": "[10,20]"}], f)

    stop_event = threading.Event()
    done_event = threading.Event()

    thread = threading.Thread(
        target=replay,
        args=(
            "examples.params.worker:ParamWorker",
            "examples.dynamic_reduce.reducer:InteractiveReducer",
            None,
            "examples.dynamic_reduce.source:SlowSource",
            par_file,
        ),
        kwargs={"port": 5010, "stop_event": stop_event, "done_event": done_event},
    )
    thread.start()
    async with aiohttp.ClientSession() as session:
        while True:
            try:
                await session.get("http://localhost:5010/api/v1/result/")
                break
            except aiohttp.client_exceptions.ClientConnectorError:
                pass
        await asyncio.sleep(1)
        st = await session.get("http://localhost:5010/api/v1/result/")
        content = await st.content.read()
        result = pickle.loads(content)[0]
        logging.warning("params in reducer is %s", result)
        assert result["params"]["int_param"].value == 0

        resp = await session.post(
            "http://localhost:5010/api/v1/parameter/int_param",
            data=b"42",
        )
        assert resp.status == 200

        await asyncio.sleep(3)
        st = await session.get("http://localhost:5010/api/v1/result/")
        content = await st.content.read()
        result = pickle.loads(content)[0]
        logging.warning("params in reducer is %s", result)
        assert result["params"]["int_param"].value == 42

    done_event.wait()

    stop_event.set()

    thread.join()
